from dataclasses import dataclass
from typing import List, Optional, Dict, Any, Union
from abc import ABC, abstractmethod

OPERATORS = None

def set_operators(operators: dict):
    global OPERATORS
    OPERATORS = operators

@dataclass
class ASTNode(ABC):
    """抽象構文木（AST）の基底クラス"""
    filename: str
    source: str
    pos: int
    end_pos: int
    comment: Optional[str]

    def __init__(self):
        """ASTNodeを初期化する"""
        self.filename = "main.yui"
        self.source = ""
        self.pos = 0
        self.end_pos = -1
        self.comment = None

    def setpos(self, source: str, pos: int, end_pos: int = -1, filename: str = "main.yui"):
        self.source = source
        self.pos = pos
        self.end_pos = end_pos
        self.filename = filename
        self.comment = None
        return self
    
    def __str__(self) -> str:
        """ノードに対応するソースコードのスニペットを返す"""
        return self.source[self.pos:self.end_pos]

    def evaluate(self, runtime: 'YuiRuntime') -> 'YuiValue': # type: ignore
        return self.visit(runtime)

    def visit(self, visitor):
        """ノードを訪問する"""
        method_name = 'visit' + self.__class__.__name__
        visit = getattr(visitor, method_name, visitor.visitASTNode)
        return visit(self)
    
    def extract(self) -> tuple:
        """ソースコード内の位置をエラー表示用の情報に変換する

        Returns:
            tuple: (line, col, snippet)
                - line: 行番号（1始まり）
                - col: 列番号（1始まり）
                - snippet: エラー行のコードスニペット
        """
        linenum = 1
        col = 1
        start = 0

        # エラー位置まで文字を辿り、行番号と列番号を計算
        for i, char in enumerate(self.source):
            if i == self.pos:
                break
            if char == '\n':
                linenum += 1
                col = 1
                start = i + 1
            else:
                col += 1

        # エラー行の終端を見つける
        end_pos = self.source.find('\n', start)
        if end_pos == -1:
            end_pos = len(self.source)
        return linenum, col, self.source[start:end_pos]


@dataclass
class ExpressionNode(ASTNode):
    """式（Expression）の基底クラス"""
    def __init__(self):
        super().__init__()

def _node(node: Any) -> ASTNode:
    if node is None: return None
    if isinstance(node, (int, float)):
        return NumberNode(node)
    if isinstance(node, str):
        return StringNode(node)
    if isinstance(node, list):
        return ArrayNode([_node(e) for e in node])
    assert isinstance(node, ASTNode)
    return node

@dataclass
class ConstNode(ExpressionNode):
    """null/boolean値を表すノード (native_value: None, True, or False)"""
    native_value: Any  # None, True, or False

    def __init__(self, value=None):
        super().__init__()
        self.native_value = value

    def visit(self, visitor):
        return visitor.visitConstNode(self)


@dataclass
class NumberNode(ExpressionNode):
    """数値リテラルを表すノード"""
    native_value: Union[int, float]

    def __init__(self, value: Union[int, float]):
        super().__init__()
        self.native_value = value

    def visit(self, visitor):
        return visitor.visitNumberNode(self)

@dataclass
class ArrayLenNode(ExpressionNode):
    """配列の長さ（|配列|）を表すノード"""
    element: ExpressionNode

    def __init__(self, element: ExpressionNode):
        super().__init__()
        self.element = element

    def visit(self, visitor):
        return visitor.visitArrayLenNode(self)

@dataclass
class MinusNode(ExpressionNode):
    """負の数（-式）を表すノード"""
    element: ExpressionNode

    def __init__(self, element: ExpressionNode):
        super().__init__()
        self.element = element

    def visit(self, visitor):
        return visitor.visitMinusNode(self)

@dataclass
class StringNode(ExpressionNode):
    """文字列リテラル（"..."）を表すノード"""
    contents: Union[str, List[Union[str, ExpressionNode]]]

    def __init__(self, contents: Union[str, List[Union[str, ExpressionNode]]]):
        super().__init__()
        self.contents = contents

    def visit(self, visitor):
        return visitor.visitStringNode(self)

@dataclass
class ArrayNode(ExpressionNode):
    """配列リテラル（[要素, ...]）を表すノード"""
    elements: List[Any]

    def __init__(self, elements: List[Any]):
        super().__init__()
        self.elements = [_node(e) for e in elements]

    def visit(self, visitor):
        return visitor.visitArrayNode(self)

@dataclass
class ObjectNode(ExpressionNode):
    """辞書リテラル（{要素, ...}）を表すノード"""
    elements: List[Any]

    def __init__(self, elements: List[Any]):
        super().__init__()
        self.elements = [_node(e) for e in elements]

    def visit(self, visitor):
        return visitor.visitObjectNode(self)

@dataclass
class NameNode(ExpressionNode):
    """変数参照を表すノード"""
    name: str

    def __init__(self, name: str):
        super().__init__()
        self.name = name

    def visit(self, visitor):
        return visitor.visitNameNode(self)
    
    def update(self, value, visitor):
        """変数の値を更新する"""
        visitor.setenv(self.name, value)

@dataclass
class GetIndexNode(ASTNode):
    """配列またはオブジェクトのインデックス取得を表すノード"""
    collection: ExpressionNode
    index_node: ExpressionNode

    def __init__(self, collection: ExpressionNode, index: ExpressionNode):
        super().__init__()
        self.collection = _node(collection)
        self.index_node = _node(index)

    def visit(self, visitor):
        return visitor.visitGetIndexNode(self)
    
    def update(self, value, visitor):
        """変数の値を更新する"""
        collection = self.collection.visit(visitor)
        index = self.index_node.visit(visitor)
        collection.set_item(index, value)

@dataclass
class BinaryNode(ASTNode):
    """二項演算子を表すノード"""
    left_node: ExpressionNode
    right_node: ExpressionNode
    operator: str
    comparative: bool

    def __init__(self, left: ExpressionNode, operator: str, right: ExpressionNode, comparative: bool = False):
        super().__init__()
        self.left_node = _node(left)
        self.operator = operator
        self.right_node = _node(right)
        self.comparative = comparative

    # def evaluate(self, runtime):
    #     raise YuiError(("error", "internal", f"🔍{self.operator} operator is not implemented"), self, runtime)

@dataclass
class FuncAppNode(ExpressionNode):
    """関数呼び出し（関数名(引数, ...)）を表すノード"""
    name: ExpressionNode
    arguments: List[ExpressionNode]
    snippet: str

    def __init__(self, name: ExpressionNode, arguments: List[ExpressionNode]):
        super().__init__()
        self.name_node = NameNode(name) if isinstance(name, str) else _node(name)
        self.arguments = [_node(arg) for arg in arguments]
        self.snippet = str(self)

    def visit(self, visitor):
        return visitor.visitFuncAppNode(self)


@dataclass
class StatementNode(ASTNode):
    """文（Statement）の基底クラス"""
    def __init__(self):
        super().__init__()

@dataclass
class AssignmentNode(StatementNode):
    """代入文（変数 = 式）を表すノード"""
    variable: NameNode
    expression: ExpressionNode

    def __init__(self, variable: NameNode, expression: ExpressionNode):
        super().__init__()
        self.variable = NameNode(variable) if isinstance(variable, str) else _node(variable)
        self.expression = _node(expression)

    def visit(self, visitor):
        return visitor.visitAssignmentNode(self)

@dataclass
class IncrementNode(StatementNode):
    """インクリメント（変数 を 増やす）を表すノード"""
    variable: NameNode

    def __init__(self, variable: NameNode):
        super().__init__()
        self.variable = NameNode(variable) if isinstance(variable, str) else _node(variable)

    def visit(self, visitor):
        return visitor.visitIncrementNode(self)

@dataclass
class DecrementNode(StatementNode):
    """デクリメント（変数 を 減らす）を表すノード"""
    variable: NameNode

    def __init__(self, variable: NameNode):
        super().__init__()
        self.variable = NameNode(variable) if isinstance(variable, str) else _node(variable)

    def visit(self, visitor):
        return visitor.visitDecrementNode(self)

@dataclass
class AppendNode(StatementNode):
    """配列への追加（変数の末尾に 値 を 追加する）を表すノード"""
    variable: NameNode
    expression: ExpressionNode

    def __init__(self, variable: NameNode, expression: ExpressionNode):
        super().__init__()
        self.variable = NameNode(variable) if isinstance(variable, str) else _node(variable)
        self.expression = _node(expression)

    def visit(self, visitor):
        return visitor.visitAppendNode(self)

@dataclass
class BlockNode(StatementNode):
    statements: List[StatementNode]
    top_level: bool

    def __init__(self, statements: List[StatementNode], top_level: bool = False):
        super().__init__()
        if isinstance(statements, StatementNode):
            self.statements = [statements]
        else:
            assert isinstance(statements, list)
            self.statements = statements
        self.top_level = top_level

    def visit(self, visitor):
        return visitor.visitBlockNode(self)

@dataclass
class IfNode(StatementNode):
    """条件分岐（もし〜ならば）を表すノード"""
    left: ExpressionNode
    operator: str  # "==", "!=", ">", "<", "!=", "in", "not in"
    right: ExpressionNode
    then_block: BlockNode
    else_block: Optional[BlockNode] = None

    def __init__(self, 
                 left: ExpressionNode, operator: str, right: ExpressionNode,
                 then_block: BlockNode, else_block: Optional[BlockNode] = None):
        super().__init__()
        self.left = _node(left)
        self.operator = OPERATORS[operator]
        self.right = _node(right)
        self.then_block = then_block
        self.else_block = else_block

    def visit(self, visitor):
        return visitor.visitIfNode(self)

@dataclass
class BreakNode(StatementNode):
    def __init__(self):
        super().__init__()

    def visit(self, visitor):
        return visitor.visitBreakNode(self)

@dataclass
class PassNode(StatementNode):
    def __init__(self, comment: Optional[str] = None):
        super().__init__()
        self.comment = comment

    def visit(self, visitor):
        return visitor.visitPassNode(self)

@dataclass
class RepeatNode(StatementNode):
    """ループ（N回 くり返す）を表すノード"""
    count: ExpressionNode
    body: BlockNode

    def __init__(self, count_node: ExpressionNode, block_node: BlockNode):
        super().__init__()
        self.count_node = _node(count_node)    
        self.block_node = block_node

    def visit(self, visitor):
        return visitor.visitRepeatNode(self)

@dataclass
class ImportNode(StatementNode):
    """ライブラリのインポート（標準ライブラリを使う）を表すノード"""
    module_name: str

    def __init__(self, module_name: str = None):
        super().__init__()
        self.module_name = NameNode(module_name) if isinstance(module_name, str) else _node(module_name)

    def visit(self, visitor):
        return visitor.visitImportNode(self)

@dataclass
class ReturnNode(StatementNode):
    """関数からの返値（式 が 答え）を表すノード"""
    expression: ExpressionNode

    def __init__(self, expression: ExpressionNode):
        super().__init__()
        self.expression = _node(expression)

    def visit(self, visitor):
        return visitor.visitReturnNode(self)

@dataclass
class FuncDefNode(StatementNode):
    """関数定義を表すノード"""
    name_node: NameNode
    parameters: List[NameNode]
    body: BlockNode

    def __init__(self, name_node: NameNode, parameters: List[NameNode], body: BlockNode):
        super().__init__()
        self.name_node = NameNode(name_node) if isinstance(name_node, str) else _node(name_node)
        self.parameters = [NameNode(param) if isinstance(param, str) else _node(param) for param in parameters]
        self.body = body

    def visit(self, visitor):
        return visitor.visitFuncDefNode(self)


@dataclass
class PrintExpressionNode(StatementNode):
    """式の出力（単独で書かれた式）を表すノード"""
    expression: ExpressionNode
    inspection: bool
    groping: bool = False

    def __init__(self, expression: ExpressionNode, inspection: bool = False, groping: bool = False):
        super().__init__()
        self.expression = _node(expression)
        self.inspection = inspection
        self.groping = groping

    def visit(self, visitor):
        return visitor.visitPrintExpressionNode(self)

@dataclass
class AssertNode(StatementNode):
    """テストケース（>>> 式 → 期待値）を表すノード"""
    test: ExpressionNode
    reference: ExpressionNode

    def __init__(self, test: ExpressionNode, reference: ExpressionNode):
        super().__init__()
        self.test = _node(test)
        self.reference = _node(reference)

    def visit(self, visitor):
        return visitor.visitAssertNode(self)





import pytest

from .yuiast import (
    ConstNode, NumberNode, StringNode,
    ArrayNode, ObjectNode,
    NameNode,
    MinusNode, ArrayLenNode, GetIndexNode,
    AssignmentNode, IncrementNode, DecrementNode, AppendNode,
    BlockNode, IfNode, RepeatNode, BreakNode,
    FuncDefNode, FuncAppNode, ReturnNode,
    PrintExpressionNode, AssertNode,
)
from .yuicoding import CodingVisitor
from .yuisyntax import load_syntax

yui_syntax = load_syntax('yui')

class TestCodeGeneration:
    """Yuiコード生成に関するテストクラス"""

    def init_coder(self):
        return CodingVisitor(syntax_json=yui_syntax)

    def test_ConstNode_null(self):
        emitter = self.init_coder()
        assert emitter.emit(ConstNode(None)) == '値なし'

    def test_ConstNode_true(self):
        emitter = self.init_coder()
        assert emitter.emit(ConstNode(True)) == '真'

    def test_ConstNode_false(self):
        emitter = self.init_coder()
        assert emitter.emit(ConstNode(False)) == '偽'

    def test_NumberNode(self):
        node = NumberNode(123)
        emitter = self.init_coder()
        assert emitter.emit(node) == '123'

    def test_StringNode(self):
        node = StringNode("hello")
        emitter = self.init_coder()
        assert emitter.emit(node) == '"hello"'

    def test_ArrayNode(self):
        node = ArrayNode([NumberNode(1), NumberNode(2), NumberNode(3)])
        emitter = self.init_coder()
        assert emitter.emit(node) == '[1,2,3]'

    def test_ObjectNode(self):
        node = ObjectNode([
            StringNode("a"), NumberNode(1),
            StringNode("b"), StringNode("two"),
        ])
        emitter = self.init_coder()
        assert emitter.emit(node) == '{"a":1,"b":"two"}'

    def test_MinusNode(self):
        node = MinusNode(NumberNode(5))
        emitter = self.init_coder()
        assert emitter.emit(node) == '-5'
    
    def test_ArrayLenNode(self):
        node = ArrayLenNode(NameNode("arr"))
        emitter = self.init_coder()
        assert emitter.emit(node) == 'arrの大きさ'
    
    def test_GetIndexNode(self):
        node = GetIndexNode(NameNode("arr"), NumberNode(2))
        emitter = self.init_coder()
        assert emitter.emit(node) == 'arr[2]'
    
    def test_AssignmentNode(self):
        node = AssignmentNode(NameNode("x"), NumberNode(10))
        emitter = self.init_coder()
        assert emitter.emit(node) == 'x=10'

    def test_IncrementNode(self):
        node = IncrementNode(NameNode("x"))
        emitter = self.init_coder()
        assert emitter.emit(node) == 'xを増やす'

    def test_DecrementNode(self):
        node = DecrementNode(NameNode("x"))
        emitter = self.init_coder()
        assert emitter.emit(node) == 'xを減らす'

    def test_AppendNode(self):
        node = AppendNode(NameNode("arr"), NumberNode(10))
        emitter = self.init_coder()
        assert emitter.emit(node) == 'arrに10を追加する'

    def test_BreakNode(self):
        node = BreakNode()
        emitter = self.init_coder()
        assert emitter.emit(node) == 'くり返しを抜ける'

    def test_ReturnNode(self):
        node = ReturnNode(NameNode("result"))
        emitter = self.init_coder()
        assert emitter.emit(node) == 'resultが答え'

    def test_PrintExpressionNode(self):
        node = PrintExpressionNode(StringNode("Hello, World!"))
        emitter = self.init_coder()
        assert emitter.emit(node) == '"Hello, World!"'

    def test_BlockNode(self):
        node = BlockNode([
            AssignmentNode(NameNode("x"), NumberNode(1)),
            AssignmentNode(NameNode("y"), NumberNode(2)),
        ], top_level=True)
        emitter = self.init_coder()
        assert emitter.emit(node) == 'x=1\ny=2'


from typing import Dict

from .yuiast import (
    ASTNode, ConstNode, NameNode,
    StringNode, NumberNode, ArrayNode, ObjectNode,
    MinusNode, ArrayLenNode,
    FuncAppNode, GetIndexNode, BinaryNode,
    AssignmentNode, IncrementNode, DecrementNode, AppendNode,
    BlockNode, PrintExpressionNode, PassNode,
    IfNode, BreakNode, RepeatNode, FuncDefNode, ReturnNode,
    AssertNode, ImportNode,
)
from .yuitypes import YuiType
from .yuisyntax import load_syntax, YuiSyntax

class CodingVisitor(YuiSyntax):

    def __init__(self, syntax_json: Dict[str, str]):
        super().__init__(syntax_json)
        self.buffer = []
        self.indent = 0
        self.just_linefeeded = False

    def emit(self, node: ASTNode) -> str:
        self.buffer = []
        self.indent = 0
        self.just_linefeeded = True
        node.visit(self)
        return ''.join(self.buffer)

    def last_char(self) -> str:
        if len(self.buffer) == 0:
            return '\n'
        return self.buffer[-1][-1]

    def linefeed(self):
        if not self.just_linefeeded:
            self.buffer.append('\n' + '  ' * self.indent)
            self.just_linefeeded = True

    def string(self, text: str):
        if '\n' in text:
            lines = text.split('\n')
            for line in lines[:-1]:
                self.string(line)
                self.linefeed()
            self.string(lines[-1])
            return
        if len(text) == 0:
            return
        if text == " " and self.last_char() == ' ':
            return # avoid consecutive spaces
        self.buffer.append(text)
        self.just_linefeeded = False

    def word_segment(self, no_spaece_if_last_chars=' \n([{'):
        if self.is_defined('word-segmenter'):
            if self.last_char() not in no_spaece_if_last_chars:
                self.string(' ')

    def terminal(self, terminal: str, if_undefined = None, linefeed_before=False):
        if terminal == 'linefeed':
            self.linefeed()
            return
        if not self.is_defined(terminal):
            return
        token = self.for_example(terminal)
        if token == "": 
            print(f"Warning: terminal '{terminal}' is empty string")
            return
        if token[0] not in ",()[]{}:\"'.": 
            # avoid unnecessary word segmentation before terminals
            self.word_segment()
        if linefeed_before:
            self.linefeed()
        self.string(token)

    def comment(self, comment: str):
        if not comment:
            return
        if self.is_defined('comment-begin') and self.is_defined('comment-end'):
            self.terminal('comment-begin')
            self.string(comment)
            self.terminal('comment-end')
        elif self.is_defined('line-comment-begin'):
            for line in comment.splitlines():
                self.terminal('line-comment-begin')
                self.string(f' {line}')
                self.linefeed()

    def expression(self, node: ASTNode):
        self.word_segment()
        node.visit(self)

    def statement(self, node: ASTNode):
        node.visit(self)
        self.comment(node.comment)

    def block(self, node: ASTNode):
        if not isinstance(node, BlockNode):
            BlockNode([node]).visit(self)
        else:
            node.visit(self)

    def escape(self, text: str) -> str:
        return text.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')

    # AST Node Visitors

    def visitASTNode(self, node: ASTNode):
        self.string(f'FIXME: {node.__class__.__name__}')

    def visitConstNode(self, node: ConstNode):
        if node.native_value is None:
            self.terminal('null')
        elif node.native_value is True:
            self.terminal('boolean-true')
        else:
            self.terminal('boolean-false')

    def visitNumberNode(self, node: NumberNode):
        self.terminal("number-begin")
        self.string(YuiType.arrayview_s(node.native_value))
        self.terminal("number-end")

    def visitStringNode(self, node: StringNode):
        self.terminal('string-begin')
        if isinstance(node.contents, str):
            self.string(self.escape(node.contents))
        else:
            for content in node.contents:
                if isinstance(content, str):
                    self.string(self.escape(content))
                else:
                    self.terminal('string-interpolation-begin')
                    content.visit(self)
                    self.terminal('string-interpolation-end')
        self.terminal('string-end')
    
    def visitNameNode(self, node: NameNode):
        self.terminal("name-begin")
        self.string(node.name)
        self.terminal("name-end")

    def visitArrayNode(self, node: ArrayNode):
        saved_buffer = self.buffer
        self.buffer = []
        self.terminal('array-begin')
        for i, element in enumerate(node.elements):
            if i > 0:
                self.terminal('array-separator')
            self.expression(element)
        self.terminal('array-end')
        content = ''.join(self.buffer)
        self.buffer = saved_buffer
        if len(content) <= 80 and '\n' not in content:
            self.string(content)
            return
        self.terminal('array-begin')
        self.indent += 1
        self.linefeed()
        for i, element in enumerate(node.elements):
            if i > 0:
                self.terminal('array-separator')
                self.linefeed()
            self.expression(element)
        self.indent -= 1
        self.linefeed()        
        self.terminal('array-end')
        
    def visitObjectNode(self, node: ObjectNode):
        saved_buffer = self.buffer
        self.buffer = []
        self.terminal('object-begin')
        for i in range(0, len(node.elements), 2):
            if i > 0:
                self.terminal('object-separator')
            key_node = node.elements[i]
            value_node = node.elements[i+1]
            self.expression(key_node)
            self.terminal('key-value-separator')
            self.expression(value_node)
        self.terminal('object-end')
        content = ''.join(self.buffer)
        self.buffer = saved_buffer
        if len(content) <= 80 and '\n' not in content:
            self.string(content)
            return
        self.terminal('object-begin')
        self.indent += 1
        self.linefeed()
        for i in range(0, len(node.elements), 2):
            if i > 0:
                self.terminal('object-separator')
                self.linefeed()
            key_node = node.elements[i]
            value_node = node.elements[i+1]
            self.expression(key_node)
            self.terminal('key-value-separator')
            self.expression(value_node)
        self.indent -= 1
        self.linefeed()
        self.terminal('object-end')

    def visitMinusNode(self, node: MinusNode):
        if self.is_defined('minus-begin'):
            self.terminal('minus-begin')
            self.expression(node.element)
            self.terminal('minus-end')
        elif self.is_defined("unary-minus"):
            self.terminal("unary-minus")
            node.element.visit(self) # avoid extra word segmenter for negative numbers
        else:
            self.visitASTNode(node)

    def visitBinaryNode(self, node: BinaryNode):
        self.expression(node.left_node)
        self.word_segment()
        self.terminal(f"binary{node.operator}")
        self.word_segment()
        self.expression(node.right_node)

    def visitArrayLenNode(self, node: ArrayLenNode):
        if self.is_defined('property-length'):
            self.expression(node.element)
            self.terminal('property-accessor')
            self.terminal('property-length')
        elif self.is_defined('unary-length'):
            self.terminal('unary-length')
            self.expression(node.element)
        elif self.is_defined('length-begin'):
            self.terminal('length-begin')
            self.expression(node.element)
            self.terminal('length-end')

    def visitGetIndexNode(self, node: GetIndexNode):
        self.terminal('array-indexer-begin')
        self.expression(node.collection)
        self.terminal('array-indexer-suffix')
        self.expression(node.index_node)
        self.terminal('array-indexer-end')
    
    def visitFuncAppNode(self, node: FuncAppNode):
        self.terminal('funcapp-begin')
        self.expression(node.name_node)
        self.terminal('funcapp-args-suffix')
        for i, arg in enumerate(node.arguments):
            if i > 0:
                self.terminal('funcapp-args-separator')
            self.expression(arg)
        self.terminal('funcapp-args-end')

    def visitAssignmentNode(self, node: AssignmentNode):
        self.terminal('assignment-begin')
        self.expression(node.variable)
        self.terminal('assignment-infix')
        self.expression(node.expression)
        self.terminal('assignment-end')
    
    def visitIncrementNode(self, node: IncrementNode):
        self.terminal('increment-begin')
        self.expression(node.variable)
        self.terminal('increment-infix')
        self.terminal('increment-end')

    def visitDecrementNode(self, node: DecrementNode):
        self.terminal('decrement-begin')
        self.expression(node.variable)
        self.terminal('decrement-infix')
        self.terminal('decrement-end')

    def visitAppendNode(self, node: AppendNode):
        self.terminal('append-begin')
        self.expression(node.variable)
        self.terminal('append-infix')
        self.expression(node.expression)
        self.terminal('append-suffix')
        self.terminal('append-end')

    def visitBreakNode(self, node: BreakNode):
        self.terminal('break')
    
    def visitPassNode(self, node: PassNode):
        # block 内で処理される
        # self.terminal('pass')
        pass

    def visitReturnNode(self, node: ReturnNode):
        if isinstance(node.expression, ASTNode):
            self.terminal('return-begin')
            self.expression(node.expression)
            self.terminal('return-end')
        else:
            self.terminal('return-none')
        
    def visitPrintExpressionNode(self, node: PrintExpressionNode):
        if node.groping:
            self.terminal('groping-begin')
            self.expression(node.expression)
            self.terminal('groping-end')
        elif node.inspection and self.is_defined('unary-inspection'):
            self.terminal('unary-inspection')
            self.expression(node.expression)
        else:
            self.terminal('print-begin')
            self.expression(node.expression)
            self.terminal('print-end')

    def visitIfNode(self, node: IfNode):
        self.terminal('if-begin')
        self.terminal('if-condition-begin')
        self.expression(node.left)
        if isinstance(node.left, BinaryNode) and node.left.comparative:
            pass
        else:
            if self.is_defined(f'if-infix{node.operator}'):
                self.terminal(f'if-infix{node.operator}')
            else:
                self.terminal('if-infix')
            self.expression(node.right)
            if self.is_defined(f'if-suffix{node.operator}'):
                self.terminal(f'if-suffix{node.operator}')
            else:
                self.terminal('if-suffix')
            self.terminal('if-condition-end')
        self.terminal('if-then')
        self.block(node.then_block)
        if node.else_block:
            if self.is_defined('if-else-if') and isinstance(node.else_block, IfNode):
                self.terminal('if-else-if', linefeed_before=True)
                self.block(node.else_block)
            else:
                self.terminal('if-else', linefeed_before=True)
                self.block(node.else_block)
        self.terminal('if-end', linefeed_before=True)

    def visitRepeatNode(self, node: RepeatNode):
        self.terminal('repeat-begin')
        self.expression(node.count_node)
        self.terminal('repeat-times')
        self.terminal('repeat-block')
        self.block(node.block_node)
        self.terminal('repeat-end', linefeed_before=True)

    def visitFuncDefNode(self, node: FuncDefNode):
        self.terminal('funcdef-begin')
        self.terminal('funcdef-name-begin')
        self.expression(node.name_node)
        self.terminal('funcdef-name-end')
        if self.is_defined('funcdef-noarg') and len(node.parameters) == 0:
            self.terminal('funcdef-noarg')
        else:
            self.terminal('funcdef-args-begin')
            for i, arg_node in enumerate(node.parameters):
                if i > 0:
                    self.terminal('funcdef-arg-separator')
                self.expression(arg_node)
            self.terminal('funcdef-args-end')
        self.terminal('funcdef-block')
        self.block(node.body)
        self.terminal('funcdef-end', linefeed_before=True)

    def visitAssertNode(self, node: AssertNode):
        self.terminal('assert-begin')
        self.expression(node.test)
        self.terminal('assert-infix')
        self.expression(node.reference)
        self.terminal('assert-end')

    def visitBlockNode(self, node: BlockNode):
        if not node.top_level:
            self.terminal('block-begin')
            self.indent += 1
            self.linefeed()

        if len(node.statements) == 0:
            self.terminal('pass')
        else: 
            for i, statement in enumerate(node.statements):
                if i > 0:
                    self.linefeed()
                statement.visit(self)
                if isinstance(statement, FuncDefNode):
                    self.linefeed()
                self.terminal("block-separator")
                if isinstance(statement, PassNode):
                    self.linefeed()
                self.comment(statement.comment)

        if not node.top_level:
            self.indent -= 1
            self.just_linefeeded = False  # indent 変化後に正しいインデントで linefeed させる
            self.terminal('block-end', linefeed_before=True)



"""
YuiRuntime.exec() を通じた実行テスト

すべてのコード実行は Runtime.exec() を通るため、ここでは exec() の
インターフェースを使ったテストを書く。
"""

import pytest
from .yuiruntime import YuiRuntime
from .yuitypes import YuiValue, YuiType, YuiError


# ─────────────────────────────────────────────────────────────────────────────
# ヘルパー
# ─────────────────────────────────────────────────────────────────────────────

# 標準ライブラリを使う宣言 (yui 構文の import-standard キーワード)
STDLIB = "標準ライブラリを使う\n"


def run(source: str, syntax: str = 'yui') -> dict:
    """source を exec() で実行し、最終環境 dict を返す"""
    rt = YuiRuntime()
    rt.exec(source, syntax, eval_mode=False)
    return rt.environments[-1]


def run_std(source: str, syntax: str = 'yui') -> dict:
    """標準ライブラリ込みで source を exec() で実行し、最終環境 dict を返す"""
    return run(STDLIB + source, syntax)


def val(env: dict, name: str):
    """環境から変数の Python native 値を取得する"""
    return YuiType.to_native(env[name])


def run_rt(source: str, syntax: str = 'yui') -> YuiRuntime:
    """標準ライブラリ込みで source を exec() で実行し、Runtime ごと返す"""
    rt = YuiRuntime()
    rt.exec(STDLIB + source, syntax, eval_mode=False)
    return rt


# ─────────────────────────────────────────────────────────────────────────────
# NativeFunction の戻り値ラップ (回帰テスト)
# ─────────────────────────────────────────────────────────────────────────────

class TestNativeFunctionWrapping:
    """
    NativeFunction.call() が plain Python 値を YuiValue でラップすることの確認。

    修正前: 和() / 積() などの整数パスが plain int を返すため、
            変数に代入後に .type 参照で AttributeError が発生し、
            アサートが ❌10 ✅10 のように偽陰性になっていた。
    修正後: NativeFunction.call() で isinstance チェックして YuiValue に統一。
    """

    def test_sum_stores_yuivalue(self):
        """和() の結果が環境に YuiValue として格納される"""
        env = run_std("x = 和(3, 4)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == 7

    def test_product_stores_yuivalue(self):
        """積() の結果が環境に YuiValue として格納される"""
        env = run_std("x = 積(3, 4)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == 12

    def test_diff_stores_yuivalue(self):
        env = run_std("x = 差(10, 3)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == 7

    def test_quotient_stores_yuivalue(self):
        env = run_std("x = 商(10, 3)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == 3

    def test_remainder_stores_yuivalue(self):
        env = run_std("x = 剰余(10, 3)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == 1

    def test_max_stores_yuivalue(self):
        env = run_std("x = 最大値(3, 7, 2)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == 7

    def test_min_stores_yuivalue(self):
        env = run_std("x = 最小値(3, 7, 2)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == 2

    def test_and_stores_yuivalue(self):
        env = run_std("x = 論理積(12, 10)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == (12 & 10)

    def test_or_stores_yuivalue(self):
        env = run_std("x = 論理和(12, 10)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == (12 | 10)

    def test_isint_stores_yuivalue(self):
        env = run_std("x = 整数判定(42)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == 1

    def test_toint_stores_yuivalue(self):
        env = run_std("x = 整数化(3.9)")
        assert isinstance(env['x'], YuiValue)
        assert val(env, 'x') == 3


# ─────────────────────────────────────────────────────────────────────────────
# アサート (>>>) が NativeFunction の戻り値で正しく動作すること (回帰テスト)
# ─────────────────────────────────────────────────────────────────────────────

class TestAssertWithNativeFunctions:
    """
    NativeFunction の結果をアサートできること。

    修正前: tested が plain int のため tested.type で AttributeError →
            except Exception で握りつぶし → ❌10 ✅10 という偽陰性エラー。
    """

    def test_sum_assert_passes(self):
        """和() の結果を直接アサートできること"""
        rt = run_rt(">>> 和(3, 4)\n7")
        assert rt.test_passed  # 失敗なら exec() が YuiError を送出する

    def test_product_assert_passes(self):
        rt = run_rt(">>> 積(3, 4)\n12")
        assert rt.test_passed

    def test_sum_via_variable_assert_passes(self):
        """Native 関数の結果を変数に代入してからアサートできること (バグの主な再現ケース)"""
        code = "x = 和(3, 4)\n>>> x\n7"
        rt = run_rt(code)
        assert rt.test_passed

    def test_chained_native_assert_passes(self):
        """Native 関数をネストした呼び出し結果をアサートできること"""
        # 積(result, 10) のような内部でネイティブ関数を使うケース
        code = "x = 和(積(2, 5), 3)\n>>> x\n13"
        rt = run_rt(code)
        assert rt.test_passed

    def test_user_function_using_native_assert(self):
        """
        Native 関数を内部で使うユーザ定義関数のアサート。
        closest_integer の失敗パターンの再現。
        """
        code = (
            "ten_times = 入力 n に対して {\n"
            "  積(n, 10) が答え\n"
            "}\n"
            ">>> ten_times(5)\n"
            "50\n"
        )
        rt = run_rt(code)
        assert rt.test_passed

    def test_user_function_chained_native_assert(self):
        """ループ内でネイティブ関数を重ねて使うユーザ定義関数のアサート"""
        code = (
            "accumulate = 入力 n に対して {\n"
            "  result = 0\n"
            "  n回くり返す {\n"
            "    result = 和(result, 1)\n"
            "  }\n"
            "  result が答え\n"
            "}\n"
            ">>> accumulate(5)\n"
            "5\n"
        )
        rt = run_rt(code)
        assert rt.test_passed

    def test_assert_failure_raises_yuierror(self):
        """アサート失敗時に YuiError が送出されること"""
        with pytest.raises(YuiError):
            run_rt(">>> 和(3, 4)\n99")


# ─────────────────────────────────────────────────────────────────────────────
# 基本実行
# ─────────────────────────────────────────────────────────────────────────────

class TestBasicExec:
    def test_assignment(self):
        assert val(run("x = 42"), 'x') == 42

    def test_string_assignment(self):
        assert val(run('s = "hello"'), 's') == "hello"

    def test_increment(self):
        assert val(run("x = 0\nxを増やす"), 'x') == 1

    def test_decrement(self):
        assert val(run("x = 5\nxを減らす"), 'x') == 4

    def test_arithmetic_add(self):
        # BinaryNode (+) は未実装のため、標準ライブラリの 和() を使う
        assert val(run_std("x = 和(3, 4)"), 'x') == 7

    def test_arithmetic_mul(self):
        # BinaryNode (*) は未実装のため、標準ライブラリの 積() を使う
        assert val(run_std("x = 積(3, 4)"), 'x') == 12

    def test_repeat(self):
        assert val(run("x = 0\n3回くり返す {\n  xを増やす\n}"), 'x') == 3

    def test_if_true(self):
        env = run("x = 10\nもし x が 10 ならば {\n  x = 99\n}")
        assert val(env, 'x') == 99

    def test_if_false(self):
        env = run("x = 5\nもし x が 10 ならば {\n  x = 99\n}")
        assert val(env, 'x') == 5

    def test_user_function(self):
        # BinaryNode (+) は未実装のため、標準ライブラリの 和() を使う
        code = (
            STDLIB +
            "add = 入力 a, b に対して {\n"
            "  和(a, b) が答え\n"
            "}\n"
            "result = add(3, 4)"
        )
        assert val(run(code), 'result') == 7

    def test_array_literal(self):
        env = run("a = [1, 2, 3]")
        assert val(env, 'a') == [1, 2, 3]

    def test_array_append(self):
        env = run("a = [1, 2]\naに3を追加する")
        assert val(env, 'a') == [1, 2, 3]

    def test_multiple_exec_shares_env(self):
        """同じ runtime で exec() を2回呼んでも環境が引き継がれること"""
        rt = YuiRuntime()
        rt.exec("x = 10", 'yui', eval_mode=False)
        # BinaryNode (+) は未実装のため、インクリメントで +5 を表現
        rt.exec("y = x\nyを増やす\nyを増やす\nyを増やす\nyを増やす\nyを増やす", 'yui', eval_mode=False)
        assert YuiType.to_native(rt.environments[-1]['y']) == 15

    def test_syntax_error_raises_yuierror(self):
        with pytest.raises(YuiError):
            run("もし もし もし")



import pytest
from .yuiparser import Source, parse
from .yuiast import ConstNode
from .yuitypes import YuiError
from .yuisyntax import load_syntax

py_syntax = load_syntax('pylike')


class TestParseStatementNode_Pylike:

    def test_Assignment(self):
        source = Source("x = 1 # コメント", syntax=py_syntax)
        assignment_node = parse("@Assignment", source, pc={})
        assert str(assignment_node) == "x = 1"

    def test_AssignmentAsStatement(self):
        source = Source("x=1 # コメント", syntax=py_syntax)
        assignment_node = parse("@Statement", source, pc={})
        assert str(assignment_node) == "x=1"

    def test_Increment(self):
        source = Source("x += 1", syntax=py_syntax)
        increment_node = parse("@Increment", source, pc={})
        assert str(increment_node) == "x += 1"

        source = Source("y += 1", syntax=py_syntax)
        increment_node = parse("@Statement", source, pc={})
        assert str(increment_node) == "y += 1"

    def test_Decrement(self):
        source = Source("x -= 1", syntax=py_syntax)
        decrement_node = parse("@Decrement", source, pc={})
        assert str(decrement_node) == "x -= 1"

        source = Source("y -= 1 # コメント", syntax=py_syntax)
        decrement_node = parse("@Statement", source, pc={})
        assert str(decrement_node) == "y -= 1"

    def test_Append(self):
        source = Source("x.append(10)", syntax=py_syntax)
        append_node = parse("@Append", source, pc={})
        assert str(append_node) == "x.append(10)"

        source = Source("x.append(10) # コメント", syntax=py_syntax)
        append_node = parse("@Statement", source, pc={})
        assert str(append_node) == "x.append(10)"

    def test_Break(self):
        source = Source("break", syntax=py_syntax)
        break_node = parse("@Break", source, pc={})
        assert str(break_node) == "break"

    def test_Return(self):
        source = Source("return 1 # コメント", syntax=py_syntax)
        return_node = parse("@Return", source, pc={})
        assert str(return_node) == "return 1"

        source = Source("return 1", syntax=py_syntax)
        return_node = parse("@Statement", source, pc={})
        assert str(return_node) == "return 1"

    def test_FuncDef(self):
        source = Source("def f(x):\n  return x\n\n", syntax=py_syntax)
        funcdef_node = parse("@FuncDef", source, pc={})
        assert str(funcdef_node) == "def f(x):\n  return x\n\n"

    def test_FuncDef_no_args(self):
        source = Source("def f():\n  return 1\n\n", syntax=py_syntax)
        funcdef_node = parse("@FuncDef", source, pc={})
        assert str(funcdef_node) == "def f():\n  return 1\n\n"

    def test_FuncDef_multiple_args(self):
        source = Source("def f(x, y):\n  return x\n\n", syntax=py_syntax)
        funcdef_node = parse("@FuncDef", source, pc={})
        assert str(funcdef_node) == "def f(x, y):\n  return x\n\n"


class TestParseBlockNode_Pylike:

    def test_TopLevel(self):
        source = Source("x = 1\ny=2", syntax=py_syntax)
        top_level_node = parse("@TopLevel", source, pc={})
        assert str(top_level_node) == "x = 1\ny=2"

    def test_If(self):
        source = Source("if x==1:\n  x=1\n\n", syntax=py_syntax)
        if_node = parse("@If", source, pc={})
        assert str(if_node) == "if x==1:\n  x=1\n\n"

    def test_IfAsStatement(self):
        source = Source("if x==1:\n  x=1\n\ny=2", syntax=py_syntax)
        if_node = parse("@Statement", source, pc={})
        assert str(if_node) == "if x==1:\n  x=1\n\n"

    def test_NestedIf(self):
        source = Source("if x==1:\n  x=0\n  if x==0:\n    x=1\n\n", syntax=py_syntax)
        if_node = parse("@If", source, pc={})
        assert str(if_node) == "if x==1:\n  x=0\n  if x==0:\n    x=1\n\n"

    def test_NestedIfAsTopLevel(self):
        source = Source("if x==1:\n  x=0\n  if x==0:\n    x=1\n\ny=2", syntax=py_syntax)
        if_node = parse("@TopLevel", source, pc={})
        assert str(if_node) == "if x==1:\n  x=0\n  if x==0:\n    x=1\n\ny=2"

    def test_Repeat(self):
        source = Source("for _ in range(3):\n  x=1\n\n", syntax=py_syntax)
        repeat_node = parse("@Repeat", source, pc={})
        assert str(repeat_node) == "for _ in range(3):\n  x=1\n\n"

    def test_Repeat_with_break(self):
        source = Source("for _ in range(10):\n  count += 1\n  if count==5:\n    break\n\n", syntax=py_syntax)
        repeat_node = parse("@Repeat", source, pc={})
        assert "break" in str(repeat_node)


class TestParseConstNode_Pylike:

    def test_null(self):
        source = Source("None", syntax=py_syntax)
        node = parse("@Boolean", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is None
        assert str(node) == "None"

    def test_true(self):
        source = Source("True", syntax=py_syntax)
        node = parse("@Boolean", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is True
        assert str(node) == "True"

    def test_false(self):
        source = Source("False", syntax=py_syntax)
        node = parse("@Boolean", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is False
        assert str(node) == "False"

    def test_null_as_term(self):
        source = Source("None", syntax=py_syntax)
        node = parse("@Term", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is None

    def test_true_as_term(self):
        source = Source("True", syntax=py_syntax)
        node = parse("@Term", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is True

    def test_false_as_term(self):
        source = Source("False", syntax=py_syntax)
        node = parse("@Term", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is False

    def test_null_in_assignment(self):
        source = Source("x = None", syntax=py_syntax)
        node = parse("@Assignment", source, pc={})
        assert str(node) == "x = None"

    def test_true_in_assignment(self):
        source = Source("x = True", syntax=py_syntax)
        node = parse("@Assignment", source, pc={})
        assert str(node) == "x = True"



import random
import json
from typing import List, Any

from .yuitypes import (
    YuiValue, YuiType, YuiError,
    TY_INT, TY_FLOAT, TY_STRING, TY_ARRAY, TY_OBJECT,
)


def standard_lib(modules: list):
    """
    標準ライブラリを環境に追加する

    以下の関数が使用可能になります：
    - 絶対値(x): 絶対値
    - 乱数(): ランダムな少数
    - 乱整数(x): 0以上x未満のランダムな整数
    - 和(x, y, ...): 要素の合計
    - 差(x, y, ...): 要素の差
    - 積(x, y, ...): 要素の積
    - 商(x, y, ...): 要素の商
    - 剰余(x, y): 剰余
    - 最大値(x, y, ...): 最大値
    - 最小値(x, y, ...): 最小値

    - 論理積(x, y, ...): ビット単位の論理積
    - 論理和(x, y, ...): ビット単位の論理和
    - 排他的論理和(x, y, ...): ビット単位の排他的論理和
    - ビット反転(x): ビット単位の反転
    - 左シフト(x, n): xをnビット左シフト
    - 右シフト(x, n): xをnビット右シフト

    - 配列化(x): 配列に変換
    - 文字列化(x): 文字列に変換
    - 少数化(x): 少数に変換
    - 整数化(x): 整数に変換
    - 整数判定(x): 整数かどうか
    - 少数判定(x): 少数かどうか
    - 文字列判定(x): 文字列かどうか
    - オブジェクト化(x): オブジェクトに変換
    - オブジェクト判定(x): オブジェクトかどうか

    Args:
        modules: (名前, 関数) のペアを追加するリスト
    """

    def check_number_of_args(nodeargs: List[Any], expected: int) -> None:
        """関数の引数の数をチェックする"""
        if expected == -1: #少なくとも一つの引数が必要
            if len(nodeargs) < 1:
                raise YuiError(("required", "arguments", f"❌{len(nodeargs)}", f"✅>0"))
            return
        if len(nodeargs) != expected:
            last = nodeargs[-1] if nodeargs else None
            raise YuiError(("expected", "arguments", f"✅{expected}", f"❌{len(nodeargs)}"), last)

    def array_to_varargs(nodeargs:list) -> list:
        """引数が配列1つの場合、その要素を展開して返す"""
        if len(nodeargs) == 1 and isinstance(nodeargs[0], YuiValue):
            return nodeargs[0].array
        return nodeargs

    def yui_abs(*nodeargs: Any) -> Any:
        """絶対値を返す"""
        check_number_of_args(nodeargs, 1)
        YuiType.NumberType.match_or_raise(nodeargs[0])
        value = YuiType.matched_native(nodeargs[0])
        return YuiValue(abs(value))
    modules.append(('📏|絶対値|abs', yui_abs))

    def yui_random(*nodeargs: Any) -> Any:
        """ランダムな整数を返す"""
        check_number_of_args(nodeargs, 0)
        return YuiValue(random.random())
    modules.append((f'🎲{TY_FLOAT}|乱数|random', yui_random))

    def yui_randint(*nodeargs: Any) -> Any:
        """0以上x未満のランダムな整数を返す"""
        check_number_of_args(nodeargs, 1)
        YuiType.IntType.match_or_raise(nodeargs[0])
        x = YuiType.matched_native(nodeargs[0])
        if x <= 0:
            raise YuiError(("error", "invalid argument", f"❌{x}", f"✅>0"))
        return YuiValue(random.randint(0, x - 1))
    modules.append((f'🎲{TY_FLOAT}|乱整数|randint', yui_randint))

    def has_float_or_raise(nodeargs: List[Any]) -> bool:
        """引数リストに少数が含まれているかどうかを判定する"""
        for nodearg in nodeargs:
            YuiType.NumberType.match_or_raise(nodearg)
            if YuiType.is_float(nodearg):
                return True
        return False

    def yui_sum(*nodeargs: Any) -> Any:
        """要素の合計を返す"""
        check_number_of_args(nodeargs, -1)
        nodeargs = array_to_varargs(nodeargs)
        if has_float_or_raise(nodeargs):
            total = float(YuiType.matched_native(nodeargs[0]))
            for nodearg in nodeargs[1:]:
                total += float(YuiType.matched_native(nodearg))
            return YuiValue(total)
        else:
            total = YuiType.matched_native(nodeargs[0])
            for nodearg in nodeargs[1:]:
                total += YuiType.matched_native(nodearg)
            return YuiValue(total)
    modules.append(('🧮|和|sum', yui_sum))

    def yui_sub(*nodeargs: Any) -> Any:
        """要素の差を返す"""
        check_number_of_args(nodeargs, -1)
        nodeargs = array_to_varargs(nodeargs)
        if has_float_or_raise(nodeargs):
            total = YuiType.matched_native(nodeargs[0])
            for nodearg in nodeargs[1:]:
                total -= YuiType.matched_native(nodearg)
            return YuiValue(total)
        else:
            total = YuiType.matched_native(nodeargs[0])
            for nodearg in nodeargs[1:]:
                total -= YuiType.matched_native(nodearg)
            return YuiValue(total)
    modules.append(('➖|差|diff', yui_sub))

    def yui_product(*nodeargs: Any) -> Any:
        """要素の積を返す"""
        check_number_of_args(nodeargs, -1)
        nodeargs = array_to_varargs(nodeargs)
        if has_float_or_raise(nodeargs):
            total = float(YuiType.matched_native(nodeargs[0]))
            for nodearg in nodeargs[1:]:
                total *= YuiType.matched_native(nodearg)
            return YuiValue(total)
        else:
            total = YuiType.matched_native(nodeargs[0])
            for nodearg in nodeargs[1:]:
                total *= YuiType.matched_native(nodearg)
            return YuiValue(total)
    modules.append(('✖️|積|product', yui_product))

    def yui_div(*nodeargs: Any) -> Any:
        """要素の商を返す"""
        check_number_of_args(nodeargs, -1)
        nodeargs = array_to_varargs(nodeargs)
        if has_float_or_raise(nodeargs):
            total = float(YuiType.matched_native(nodeargs[0]))
            for nodearg in nodeargs[1:]:
                d = float(YuiType.matched_native(nodearg))
                if d == 0.0:
                    raise YuiError((f"error", "division by zero", f"❌{d}"), nodearg)
                total /= d
            return YuiValue(total)
        else:
            total = YuiType.matched_native(nodeargs[0])
            for nodearg in nodeargs[1:]:
                d = YuiType.matched_native(nodearg)
                if d == 0:
                    raise YuiError((f"error", "division by zero", f"❌{d}"), nodearg)
                total //= d
            return YuiValue(total)
    modules.append(('✂️|商|quotient', yui_div))

    def yui_mod(*nodeargs: Any) -> Any:
        """剰余を返す"""
        check_number_of_args(nodeargs, -1)
        nodeargs = array_to_varargs(nodeargs)
        if has_float_or_raise(nodeargs):
            total = float(YuiType.matched_native(nodeargs[0]))
            for nodearg in nodeargs[1:]:
                d = float(YuiType.matched_native(nodearg))
                if d == 0.0:
                    raise YuiError((f"error", "division by zero", f"❌{d}"), nodearg)
                total %= d
            return YuiValue(total)
        else:
            total = YuiType.matched_native(nodeargs[0])
            for nodearg in nodeargs[1:]:
                d = YuiType.matched_native(nodearg)
                if d == 0:
                    raise YuiError((f"error", "division by zero", f"❌{d}"), nodearg)
                total %= d
            return YuiValue(total)
    modules.append(('🍕|剰余|remainder', yui_mod))

    def yui_and(*nodeargs: Any) -> YuiValue:
        """論理積を返す"""
        total = YuiType.matched_native(nodeargs[0])
        for nodearg in nodeargs[1:]:
            total &= YuiType.matched_native(nodearg)
        return YuiValue(total)
    modules.append(('💡✖️|論理積|and', yui_and))

    def yui_or(*nodeargs: Any) -> YuiValue:
        """論理和を返す"""
        total = YuiType.matched_native(nodeargs[0])
        for nodearg in nodeargs[1:]:
            total |= YuiType.matched_native(nodearg)
        return YuiValue(total)
    modules.append(('💡➕|論理和|or', yui_or))

    def yui_xor(*nodeargs: Any) -> YuiValue:
        """排他的論理和を返す"""
        total = YuiType.matched_native(nodeargs[0])
        for nodearg in nodeargs[1:]:
            total ^= YuiType.matched_native(nodearg)
        return YuiValue(total)
    modules.append(('💡🔀|排他的論理和|xor', yui_xor))

    def yui_not(*nodeargs: Any) -> YuiValue:
        """ビット反転を返す"""
        check_number_of_args(nodeargs, 1)
        return YuiValue(~(YuiType.matched_native(nodeargs[0])))
    modules.append(('💡🔄|ビット反転|not', yui_not))

    def yui_left_shift(*nodeargs: Any) -> YuiValue:
        """左シフトを返す"""
        check_number_of_args(nodeargs, 2)
        return YuiValue(YuiType.matched_native(nodeargs[0]) << YuiType.matched_native(nodeargs[1]))
    modules.append(('💡⬅️|左シフト|shl', yui_left_shift))

    def yui_right_shift(*nodeargs: Any) -> YuiValue:
        """右シフトを返す"""
        check_number_of_args(nodeargs, 2)
        return YuiValue(YuiType.matched_native(nodeargs[0]) >> YuiType.matched_native(nodeargs[1]))
    modules.append(('💡➡️|右シフト|shr', yui_right_shift))

    def yui_max(*nodeargs: Any) -> Any:
        """最大値を返す"""
        check_number_of_args(nodeargs, -1)
        nodeargs = array_to_varargs(nodeargs)
        result = max(YuiType.matched_native(nodearg) for nodearg in nodeargs)
        return YuiValue(int(result) if isinstance(result, int) else result)
    modules.append(('👑|最大値|max', yui_max))

    def yui_min(*nodeargs: Any) -> Any:
        """最小値を返す"""
        check_number_of_args(nodeargs, -1)
        nodeargs = array_to_varargs(nodeargs)
        result = min(YuiType.matched_native(nodearg) for nodearg in nodeargs)
        return YuiValue(int(result) if isinstance(result, int) else result)
    modules.append(('🐜|最小値|min', yui_min))

    def yui_isint(*nodeargs: Any) -> YuiValue:
        """整数か判定する"""
        check_number_of_args(nodeargs, 1)
        return YuiValue.TrueValue if YuiType.is_int(nodeargs[0]) else YuiValue.FalseValue
    modules.append((f'{TY_INT}❓|整数判定|isint', yui_isint))

    def yui_toint(*nodeargs: Any) -> YuiValue:
        """整数化する"""
        check_number_of_args(nodeargs, 1)
        return YuiValue(int(YuiType.matched_native(nodeargs[0])))
    modules.append((f'{TY_INT}|整数化|toint', yui_toint))

    def yui_isfloat(*nodeargs: Any) -> YuiValue:
        """小数か判定する"""
        check_number_of_args(nodeargs, 1)
        return YuiValue.TrueValue if YuiType.is_float(nodeargs[0]) else YuiValue.FalseValue
    modules.append((f'{TY_FLOAT}❓|少数判定|isfloat', yui_isfloat))

    def yui_tofloat(*nodeargs: Any) -> Any:
        """小数化する"""
        check_number_of_args(nodeargs, 1)
        value = nodeargs[0]
        if YuiType.is_string(value):
            string_value = YuiType.matched_native(value)
            try:
                return YuiValue(float(string_value))
            except ValueError:
                raise YuiError((f"error", "conversion", f"❌{string_value}"))
        value = float(YuiType.matched_native(nodeargs[0]))
        return YuiValue(value)
    modules.append((f'{TY_FLOAT}|少数化|tofloat', yui_tofloat))

    def yui_isstring(*nodeargs: Any) -> YuiValue:
        """文字列か判定する"""
        check_number_of_args(nodeargs, 1)
        return YuiValue.TrueValue if YuiType.is_string(nodeargs[0]) else YuiValue.FalseValue
    modules.append((f'{TY_STRING}❓|文字列判定|isstring', yui_isstring))

    def yui_tostring(*nodeargs: Any) -> Any:
        """文字列に変換する"""
        check_number_of_args(nodeargs, 1)
        if YuiType.is_float(nodeargs[0]):
            v = YuiType.matched_native(nodeargs[0])
            return YuiValue(f"{v:.6f}")
        return YuiValue(str(nodeargs[0]))
    modules.append((f'{TY_STRING}|文字列化|tostring', yui_tostring))

    def yui_isobject(*nodeargs: Any) -> YuiValue:
        """オブジェクトか判定する"""
        check_number_of_args(nodeargs, 1)
        return YuiValue.TrueValue if YuiType.is_object(nodeargs[0]) else YuiValue.FalseValue
    modules.append((f'{TY_OBJECT}❓|オブジェクト判定|isobject', yui_isobject))

    def yui_toobject(*nodeargs: Any) -> Any:
        """オブジェクトに変換する"""
        check_number_of_args(nodeargs, 1)
        if YuiType.is_object(nodeargs[0]):
            return nodeargs[0]
        if YuiType.is_string(nodeargs[0]):
            s = YuiType.matched_native(nodeargs[0])
            if s.startswith('{'):
                try:
                    obj = json.loads(s)
                    return YuiValue(obj)
                except json.JSONDecodeError:
                    pass
        return YuiValue({})
    modules.append((f'{TY_OBJECT}|オブジェクト化|toobject', yui_toobject))

    def yui_isarray(*nodeargs: Any) -> YuiValue:
        """配列か判定する"""
        check_number_of_args(nodeargs, 1)
        return YuiValue.TrueValue if YuiType.is_array(nodeargs[0]) else YuiValue.FalseValue
    modules.append((f'{TY_ARRAY}❓|配列判定|isarray', yui_isarray))

    def yui_toarray(*nodeargs: Any) -> Any:
        """配列に変換する"""
        check_number_of_args(nodeargs, 1)
        value = nodeargs[0]
        if isinstance(value, YuiValue):
            _ = value.arrayview  # elements を先に確定させてから native_value を破棄
            value.native_value = None
            return value
        return YuiValue(YuiType.matched_native(value))
    modules.append((f'{TY_ARRAY}|配列化|toarray', yui_toarray))



import pytest
from .yuiparser import Source, parse
from .yuitypes import YuiError
from .yuisyntax import load_syntax

yui_syntax = load_syntax('yui')


class TestSource:
    """式の評価に関するテストクラス"""

    def test_is_defined(self):
        source = Source('', syntax=yui_syntax)
        source.update_syntax(**{
            "line-feed": r"\n",
            "whitespace": " ",
            "whitespaces": r"[ \t\r]+",
        })
        assert source.is_defined("linefeed")
        assert source.is_defined("comment-begin") == False

    def test_consume(self):
        source = Source("X  abc", pos=1)
        assert source.is_match("whitespaces", lskip_ws=False) == True
        assert source.pos == 3

        source = Source("X  abc", pos=1)
        assert source.is_match("whitespace", lskip_ws=False) == True
        assert source.pos == 2

        source = Source("Xabc", pos=1)
        assert source.is_match("whitespaces", lskip_ws=False) == False
        assert source.pos == 1
        with pytest.raises(YuiError) as excinfo:
            source = Source("Xabc", pos=1)
            source.try_match("whitespace", lskip_ws=False)
            assert source.pos == 1
        assert "expected" in str(excinfo.value)
        assert excinfo.value.error_node.pos == 1
        assert excinfo.value.error_node.end_pos == 2

    def test_consume_until(self):
        source = Source("X    \nabc", pos=1)
        assert source.consume_until("linefeed") == True
        assert source.pos == 5
        assert source.source[source.pos] == '\n'

    def test_capture_indent(self):
        source = Source("X\n    x=1\nabc", pos=3)
        source.consume_until("linefeed")
        assert source.capture_indent() == "    "

    def test_skip_whitespaces_comments(self):
        source = Source("X=1  #hoge   \ny=1", pos=3)
        source.skip_whitespaces_and_comments()
        assert source.source[source.pos] == '\n'

        source = Source("X=1  #hoge   \ny=1", pos=3)
        source.skip_whitespaces_and_comments(include_linefeed=True)
        assert source.source[source.pos] == 'y'

        source = Source("X =  (*optional*) 1  #hoge   \ny=1", pos=3)
        source.update_syntax(**{
            "comment-begin": r"\(\*",
            "comment-end": r"\*\)",
        })
        source.skip_whitespaces_and_comments(include_linefeed=True)
        assert source.source[source.pos] == '1'


class TestParseExpressionNode:

    def test_Number(self):
        source = Source("01234 #コメント", pos=1)

        number_node = parse("@Number", source, pc={})
        assert str(number_node) == '1234'

        source = Source("0123.12 #コメント", pos=1)
        number_node = parse("@Number", source, pc={})
        assert str(number_node) == '123.12'

        with pytest.raises(YuiError) as excinfo:
            source = Source("xxx")
            number_node = parse("@Number", source, pc={}, BK=True)
        assert "expected" in str(excinfo.value)
        assert excinfo.value.error_node.pos == 0
        assert excinfo.value.error_node.end_pos == 1

    def test_NumberAsExpression(self):
        source = Source("01234 #コメント", pos=1)

        number_node = parse("@Expression", source, pc={})
        assert str(number_node) == '1234'

        source = Source("0123.12 #コメント", pos=1)
        number_node = parse("@Expression", source, pc={})
        assert str(number_node) == '123.12'

    def test_String(self):
        source = Source('"ABC"', pos=0)
        string_node = parse("@String", source, pc={})
        assert str(string_node) == '"ABC"'

        source = Source('"AB{1}C"', pos=0)
        string_node = parse("@String", source, pc={})
        assert str(string_node) == '"AB{1}C"'

    def test_StringAsExpression(self):
        source = Source('"ABC"', pos=0)
        string_node = parse("@Expression", source, pc={})
        assert str(string_node) == '"ABC"'

        source = Source('"AB{1}C"', pos=0)
        string_node = parse("@Expression", source, pc={})
        assert str(string_node) == '"AB{1}C"'

    def test_String_wrong_token(self):
        with pytest.raises(YuiError) as excinfo:
            source = Source("'A'", pos=0)
            string_node = parse("@Expression", source, pc={})
        assert "wrong" in str(excinfo.value)

    def test_Array(self):
        source = Source('[1,2]', pos=0)
        array_node = parse("@Array", source, pc={})
        assert str(array_node) == '[1,2]'

        source = Source('[\n1,\n2\n]', pos=0)
        array_node = parse("@Array", source, pc={})
        assert str(array_node) == '[\n1,\n2\n]'

    def test_Object(self):
        source = Source('{"A": 1}', pos=0)
        object_node = parse("@Object", source, pc={})
        assert str(object_node) == '{"A": 1}'

    def test_Name(self):
        source = Source("x_1 = 1")
        name_node = parse("@Name", source, pc={})
        assert str(name_node) == "x_1"

    def test_GetArrayIndex(self):
        source = Source("x[0]")
        name_node = parse("@Expression", source, pc={})
        assert str(name_node) == "x[0]"

    def test_FuncAppAsPrimary(self):
        source = Source("x(0)")
        name_node = parse("@Primary", source, pc={})
        assert str(name_node) == "x(0)"

        source = Source("x()")
        name_node = parse("@Primary", source, pc={})
        assert str(name_node) == "x()"

        source = Source("x(1, 2)")
        name_node = parse("@Primary", source, pc={})
        assert str(name_node) == "x(1, 2)"

    def test_FuncAppAsExpression(self):
        source = Source("x(0)")
        name_node = parse("@Expression", source, pc={})
        assert str(name_node) == "x(0)"

        source = Source("x()")
        name_node = parse("@Expression", source, pc={})
        assert str(name_node) == "x()"

        source = Source("x(1, 2)")
        name_node = parse("@Expression", source, pc={})
        assert str(name_node) == "x(1, 2)"



import pytest
from .yuitypes import YuiValue, YuiType, YuiError
from .yuistdlib import standard_lib


@pytest.fixture(scope="module")
def stdlib():
    """標準ライブラリ関数を英語名 → 関数の dict として返す"""
    modules = []
    standard_lib(modules)
    return {name.split('|')[-1]: func for name, func in modules}


def v(x):
    """Python値 → YuiValue"""
    return YuiValue(x)


def n(result):
    """YuiValue/native → Python値"""
    return YuiType.to_native(result)


# ─────────────────────────────────────────────────────────────────────────────
# 算術
# ─────────────────────────────────────────────────────────────────────────────

class TestAbs:
    def test_positive_int(self, stdlib):    assert n(stdlib['abs'](v(5))) == 5
    def test_negative_int(self, stdlib):    assert n(stdlib['abs'](v(-3))) == 3
    def test_zero(self, stdlib):            assert n(stdlib['abs'](v(0))) == 0
    def test_positive_float(self, stdlib):  assert abs(n(stdlib['abs'](v(1.5))) - 1.5) < 1e-6
    def test_negative_float(self, stdlib):  assert abs(n(stdlib['abs'](v(-2.5))) - 2.5) < 1e-6

    def test_no_args(self, stdlib):
        with pytest.raises(YuiError): stdlib['abs']()
    def test_too_many_args(self, stdlib):
        with pytest.raises(YuiError): stdlib['abs'](v(1), v(2))
    def test_wrong_type(self, stdlib):
        with pytest.raises(YuiError): stdlib['abs'](v("hello"))


class TestSum:
    def test_two_ints(self, stdlib):    assert n(stdlib['sum'](v(3), v(4))) == 7
    def test_three_ints(self, stdlib):  assert n(stdlib['sum'](v(1), v(2), v(3))) == 6
    def test_negative(self, stdlib):    assert n(stdlib['sum'](v(-1), v(2))) == 1
    def test_floats(self, stdlib):      assert abs(n(stdlib['sum'](v(1.5), v(2.5))) - 4.0) < 1e-6
    def test_mixed(self, stdlib):       assert abs(n(stdlib['sum'](v(1), v(2.5))) - 3.5) < 1e-6

    def test_no_args(self, stdlib):
        with pytest.raises(YuiError): stdlib['sum']()


class TestDiff:
    def test_ints(self, stdlib):    assert n(stdlib['diff'](v(10), v(3))) == 7
    def test_negative(self, stdlib): assert n(stdlib['diff'](v(3), v(10))) == -7
    def test_floats(self, stdlib):  assert abs(n(stdlib['diff'](v(5.0), v(2.0))) - 3.0) < 1e-6


class TestProduct:
    def test_two_ints(self, stdlib):   assert n(stdlib['product'](v(3), v(4))) == 12
    def test_three_ints(self, stdlib): assert n(stdlib['product'](v(2), v(3), v(4))) == 24
    def test_floats(self, stdlib):     assert abs(n(stdlib['product'](v(2.0), v(3.0))) - 6.0) < 1e-6
    def test_zero(self, stdlib):       assert n(stdlib['product'](v(5), v(0))) == 0


class TestQuotient:
    def test_int_division(self, stdlib): assert n(stdlib['quotient'](v(10), v(3))) == 3
    def test_exact(self, stdlib):        assert n(stdlib['quotient'](v(10), v(2))) == 5
    def test_float(self, stdlib):        assert abs(n(stdlib['quotient'](v(10.0), v(4.0))) - 2.5) < 1e-6

    def test_zero_division_int(self, stdlib):
        with pytest.raises(YuiError): stdlib['quotient'](v(10), v(0))
    def test_zero_division_float(self, stdlib):
        with pytest.raises(YuiError): stdlib['quotient'](v(10.0), v(0.0))


class TestRemainder:
    def test_ints(self, stdlib):      assert n(stdlib['remainder'](v(10), v(3))) == 1
    def test_no_remainder(self, stdlib): assert n(stdlib['remainder'](v(9), v(3))) == 0
    def test_floats(self, stdlib):    assert abs(n(stdlib['remainder'](v(5.5), v(2.0))) - 1.5) < 1e-6

    def test_zero_division(self, stdlib):
        with pytest.raises(YuiError): stdlib['remainder'](v(10), v(0))


class TestMaxMin:
    def test_max_ints(self, stdlib):    assert n(stdlib['max'](v(3), v(1), v(4), v(1), v(5))) == 5
    def test_max_floats(self, stdlib):  assert abs(n(stdlib['max'](v(1.5), v(2.5))) - 2.5) < 1e-6
    def test_max_negative(self, stdlib): assert n(stdlib['max'](v(-3), v(-1))) == -1
    def test_min_ints(self, stdlib):    assert n(stdlib['min'](v(3), v(1), v(4))) == 1
    def test_min_floats(self, stdlib):  assert abs(n(stdlib['min'](v(1.5), v(0.5))) - 0.5) < 1e-6
    def test_min_negative(self, stdlib): assert n(stdlib['min'](v(-3), v(-1))) == -3

    def test_max_no_args(self, stdlib):
        with pytest.raises((YuiError, Exception)): stdlib['max']()
    def test_min_no_args(self, stdlib):
        with pytest.raises((YuiError, Exception)): stdlib['min']()


# ─────────────────────────────────────────────────────────────────────────────
# 乱数
# ─────────────────────────────────────────────────────────────────────────────

class TestRandom:
    def test_is_float(self, stdlib):
        assert YuiType.is_float(stdlib['random']())

    def test_range(self, stdlib):
        for _ in range(20):
            val = n(stdlib['random']())
            assert 0.0 <= val < 1.0

    def test_randint_is_int(self, stdlib):
        assert YuiType.is_int(stdlib['randint'](v(10)))

    def test_randint_range(self, stdlib):
        for _ in range(20):
            val = n(stdlib['randint'](v(10)))
            assert 0 <= val < 10

    def test_randint_one(self, stdlib):
        # 上限1 → 必ず0
        assert n(stdlib['randint'](v(1))) == 0

    def test_randint_zero_raises(self, stdlib):
        with pytest.raises(YuiError): stdlib['randint'](v(0))
    def test_randint_negative_raises(self, stdlib):
        with pytest.raises(YuiError): stdlib['randint'](v(-5))


# ─────────────────────────────────────────────────────────────────────────────
# ビット演算
# ─────────────────────────────────────────────────────────────────────────────

class TestBitwise:
    def test_and(self, stdlib):   assert n(stdlib['and'](v(0b1010), v(0b1100))) == 0b1000
    def test_and_zero(self, stdlib): assert n(stdlib['and'](v(0b1111), v(0b0000))) == 0

    def test_or(self, stdlib):    assert n(stdlib['or'](v(0b1010), v(0b1100))) == 0b1110
    def test_or_identity(self, stdlib): assert n(stdlib['or'](v(0b1010), v(0))) == 0b1010

    def test_xor(self, stdlib):   assert n(stdlib['xor'](v(0b1010), v(0b1100))) == 0b0110
    def test_xor_self(self, stdlib): assert n(stdlib['xor'](v(0b1010), v(0b1010))) == 0

    def test_not_zero(self, stdlib):      assert n(stdlib['not'](v(0))) == -1
    def test_not_minus_one(self, stdlib): assert n(stdlib['not'](v(-1))) == 0
    def test_not_positive(self, stdlib):  assert n(stdlib['not'](v(1))) == -2

    def test_shl(self, stdlib):     assert n(stdlib['shl'](v(1), v(3))) == 8
    def test_shl_zero(self, stdlib): assert n(stdlib['shl'](v(5), v(0))) == 5

    def test_shr(self, stdlib):     assert n(stdlib['shr'](v(8), v(3))) == 1
    def test_shr_zero(self, stdlib): assert n(stdlib['shr'](v(5), v(0))) == 5


# ─────────────────────────────────────────────────────────────────────────────
# 型判定
# ─────────────────────────────────────────────────────────────────────────────

class TestTypePredicates:
    # isint
    def test_isint_int(self, stdlib):    assert n(stdlib['isint'](v(42))) == 1
    def test_isint_float(self, stdlib):  assert n(stdlib['isint'](v(42.0))) == 0
    def test_isint_string(self, stdlib): assert n(stdlib['isint'](v("hello"))) == 0

    # isfloat
    def test_isfloat_float(self, stdlib):  assert n(stdlib['isfloat'](v(3.14))) == 1
    def test_isfloat_int(self, stdlib):    assert n(stdlib['isfloat'](v(3))) == 0
    def test_isfloat_string(self, stdlib): assert n(stdlib['isfloat'](v("3.14"))) == 0

    # isstring
    def test_isstring_string(self, stdlib):  assert n(stdlib['isstring'](v("hello"))) == 1
    def test_isstring_int(self, stdlib):     assert n(stdlib['isstring'](v(42))) == 0
    def test_isstring_float(self, stdlib):   assert n(stdlib['isstring'](v(3.14))) == 0

    # isarray
    def test_isarray_array(self, stdlib): assert n(stdlib['isarray'](v([1, 2, 3]))) == 1
    def test_isarray_int(self, stdlib):   assert n(stdlib['isarray'](v(42))) == 0

    # isobject
    def test_isobject_object(self, stdlib): assert n(stdlib['isobject'](v({"a": 1}))) == 1
    def test_isobject_int(self, stdlib):    assert n(stdlib['isobject'](v(42))) == 0
    def test_isobject_array(self, stdlib):  assert n(stdlib['isobject'](v([1, 2]))) == 0


# ─────────────────────────────────────────────────────────────────────────────
# 型変換
# ─────────────────────────────────────────────────────────────────────────────

class TestTypeConversions:
    # toint
    def test_toint_from_float_truncates(self, stdlib): assert n(stdlib['toint'](v(3.9))) == 3
    def test_toint_from_negative_float(self, stdlib):  assert n(stdlib['toint'](v(-3.9))) == -3
    def test_toint_from_int(self, stdlib):             assert n(stdlib['toint'](v(5))) == 5

    # tofloat
    def test_tofloat_from_int(self, stdlib):
        result = stdlib['tofloat'](v(3))
        assert YuiType.is_float(result)
        assert abs(n(result) - 3.0) < 1e-6
    def test_tofloat_from_float(self, stdlib):
        assert abs(n(stdlib['tofloat'](v(3.14))) - 3.14) < 1e-6
    def test_tofloat_from_string(self, stdlib):
        result = stdlib['tofloat'](v("2.5"))
        assert abs(n(result) - 2.5) < 1e-6

    # tostring
    def test_tostring_from_int(self, stdlib):   assert n(stdlib['tostring'](v(42))) == "42"
    def test_tostring_from_float(self, stdlib):
        assert n(stdlib['tostring'](v(3.14))) == "3.140000"
    def test_tostring_from_negative_int(self, stdlib): assert n(stdlib['tostring'](v(-5))) == "-5"

    # toarray
    def test_toarray_from_array(self, stdlib):
        result = stdlib['toarray'](v([1, 2, 3]))
        assert YuiType.is_array(result)
        assert n(result) == [1, 2, 3]

    # toobject
    def test_toobject_from_object(self, stdlib):
        result = stdlib['toobject'](v({"x": 1}))
        assert YuiType.is_object(result)



# Source
from dataclasses import dataclass
from typing import List, Optional, Dict, Any, Union
from types import FunctionType
from abc import ABC, abstractmethod
from typing import Union, Any, List
import re
import json
import os

def get_example_from_pattern(pattern: str) -> str:
    """正規表現パターンからそのパターンにマッチする文字列の例（最初の例）を取得する"""
    # エスケープされた文字を一時的に置換
    original_pattern = pattern
    ESC = [
        (r'\|', '▁｜'), (r'\[', '▁［'), (r'\]', '▁］'),
        (r'\(', '▁（'), (r'\)', '▁）'), (r'\*', '▁＊'), (r'\?', '▁？'),
        (r'\+', '▁＋'), (r'+', ''), (r'*', '?')
    ]
    for a, b in ESC:
        pattern = pattern.replace(a, b)
    #print(f"@pattern: `{original_pattern}` -> `{pattern}`")  # デバッグ用   
    processed = ''
    while len(pattern) > 0:
        s_pos = pattern.find('(') # シングルループだけ対応
        if s_pos == -1:
            processed += get_example_from_pattern_inner(pattern)
            break
        e_pos = pattern.find(')', s_pos+1)
        processed += get_example_from_pattern_inner(pattern[:s_pos])
        inner = pattern[s_pos+1:e_pos]
        pattern = pattern[e_pos+1:]
        if pattern.startswith('?'):
            pattern = pattern[1:]
        else:
            processed += get_example_from_pattern_inner(inner)
    #print(f"@processed: `{pattern}` -> `{processed}`")  # デバッグ用
    # 置換した文字を元に戻す
    ESC2 = [
        ('▁｜', r'|'), ('▁［', r'['), ('▁］', r']'), ('▁（', r'('),
        ('▁）', r')'), ('▁？', r'?'), ('▁＋', r'+'), ('▁＊', r'*'),
    ]
    for a, b in ESC2:
        processed = processed.replace(a, b)
    assert '▁' not in processed, f"Unprocessed escape sequences remain in `{original_pattern}`: `{processed}`"
    return processed

def split_heading_char(s: str):
    if s.startswith("\\"):
        # エスケープシーケンスの処理
        remaining = s[2:]
        if s.startswith('\\', 1):  # バックスラッシュ
            heading_char = '\\'
        elif s.startswith('s', 1):  # 空白文字
            heading_char = ' '
        elif s.startswith('t', 1):  # タブ
            heading_char = '\t'
        elif s.startswith('n', 1):  # 改行
            heading_char = '\n'
        elif s.startswith('r', 1):  # キャリッジリターン
            heading_char = '\r'
        elif s.startswith('d', 1):  # 数字
            heading_char = '1'
        elif s.startswith('w', 1):  # 単語文字
            heading_char = 'a'
        else:
            heading_char = s[1]
    elif s.startswith('▁', 0):
        heading_char = s[0:2]
        remaining = s[2:]
    elif s.startswith('\uFE0F', 1) or s.startswith('\u200D', 1): 
        #合字や絵文字のバリエーションセレクタを考慮
        heading_char = s[0:2]
        remaining = s[2:]
    else:
        heading_char = s[0]
        remaining = s[1:]
    if remaining.startswith("?"):
        return '', remaining[1:]
    return heading_char, remaining

def get_example_from_pattern_inner(pattern: str)-> str:
    if pattern == "":
        return ""
    # 選択肢（|）の処理：最初の選択肢を使用
    if "|" in pattern:
        pattern = pattern.split("|")[0]
    # 文字クラス [abc] の処理 
    if pattern.startswith("["):
        end_pos = pattern.find("]")
        if pattern.startswith('?', end_pos + 1):
            return get_example_from_pattern_inner(pattern[end_pos+2:])
        heading_char, _ = split_heading_char(pattern[1:end_pos])
        return heading_char + get_example_from_pattern_inner(pattern[end_pos+1:])
    heading_char, remaining = split_heading_char(pattern)
    return heading_char + get_example_from_pattern_inner(remaining)

def is_all_alnum(s: str) -> bool:
    for ch in s:
        if not (('a' <= ch <= 'z') or ('A' <= ch <= 'Z') or ('0' <= ch <= '9') or (ch == '_')):
            return False
    return True

def extract_identifiers(text):
    identifiers = []
    
    # 識別子のパターン: アルファベットまたはアンダースコアで始まり、
    # その後に英数字とアンダースコアが続く
    identifier_pattern = r'[^\s\]\[\(\)"]+'
    
    # 1. 改行と= の間（==は除外）
    pattern1 = rf'\n\s*({identifier_pattern})\s*=(?!=)'
    matches1 = re.findall(pattern1, text)
    identifiers.extend(matches1)
    
    # 2. 関数名のパターン
    pattern2 = rf'({identifier_pattern})\s*[\(]'
    matches2 = re.findall(pattern2, text)
    identifiers.extend(matches2)
    
    def has_unicode(s):
        for ch in s:
            if ord(ch) > 127:
                return True
        return False

    # Unicode文字を含む文字列のみ
    identifiers = list(set(id for id in identifiers if has_unicode(id)))
    # print(f"@Extracted identifiers: {identifiers}")  
    return list(set(identifiers))

DEFAULT_SYNTAX_JSON = {
    "whitespace": "[ \\t\\r　]",
    "whitespaces": "[ \\t\\r　]+",
    "linefeed": "[\\n]",
    "line-comment-begin": "[#＃]",

    "number-first-char": "[0-9]",
    "number-chars": "[0-9]*",
    "number-dot-char": "[\\.]",

    "name-first-char": "[A-Za-z_]",
    "name-chars": "[A-Za-z0-9_]*",
    
    "string-begin": "\"",
    "string-end": "\"",
    "string-escape": "\\\\",
    "string-interpolation-begin": "\\{",
    "string-interpolation-end": "\\}",
#    "string-content-end": "\\\\|\\{|\\\"",

    "array-begin": "\\[",
    "array-end": "\\]",
    "array-separator": ",",

    "object-begin": "\\{",
    "object-end": "\\}",
    "object-separator": ",",
    "key-value-separator": ":",

    "array-indexer-suffix": "\\[",
    "array-indexer-end": "\\]",
    
    "funcapp-args-suffix": "\\(",
    "funcapp-args-end": "\\)",
    "funcapp-args-separator": ",",
}

def load_syntax(filepath: Optional[str] = None) -> Dict[str, str]:
    """JSON文法ファイルから終端記号をロードする"""
    if filepath is None:
        filepath = "yui"

    if not os.path.isfile(filepath):
        current_dir = os.path.dirname(os.path.abspath(__file__))
        syntax_dir = os.path.join(current_dir, 'syntax')
        if not filepath.endswith('.json'):
            filepath = f"{filepath}.json"
        new_filepath = os.path.join(syntax_dir, filepath)
        if os.path.isfile(new_filepath):
            filepath = new_filepath

    with open(filepath, 'r', encoding='utf-8') as f:
        terminals = json.load(f)
    
    for key, pattern in DEFAULT_SYNTAX_JSON.items():
        if key not in terminals:
            terminals[key] = pattern

    if 'string-content-end' not in terminals:
        escape = terminals.get('string-escape', r'\\')
        interpolation = terminals.get('string-interpolation-begin', r'\{')
        string_end = terminals.get('string-end', r'\"')
        terminals['string-content-end'] = f"{escape}|{interpolation}|{string_end}"

    if 'identifiers' not in terminals:
        identifiers = extract_identifiers(json.dumps(terminals))
        terminals['identifiers'] = identifiers

    # if 'keywords' not in terminals:
    #     keywords = set()
    #     for key, pattern in terminals.items():
    #         word = get_example_from_pattern(pattern).strip()
    #         if is_all_alnum(word):
    #             if word not in keywords:
    #                 keywords.add(word)
    #     terminals['keywords'] = keywords        
    return terminals

_DEFAULT_SYNTAX_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'syntax')


def list_syntax_names(syntax_dir: Optional[str] = None) -> List[str]:
    """syntax ディレクトリにある syntax 名の一覧を返す（.json 拡張子なし、ソート済み）。"""
    d = syntax_dir or _DEFAULT_SYNTAX_DIR
    return sorted(f[:-5] for f in os.listdir(d) if f.endswith('.json'))


def find_matching_syntaxes(sources: Dict[str, str], syntax_dir: Optional[str] = None) -> List[tuple]:
    """
    利用可能な全 syntax に対して sources のソースコードを解析し、成否を返す。

    Args:
        sources:    {filename: source_code} の辞書
        syntax_dir: syntax JSON ファイルを探すディレクトリ（None なら組み込みディレクトリ）

    Returns:
        [(syntax_name, matched: bool, status_message: str), ...] のリスト
    """
    from .yuiparser import YuiParser  # 循環インポートを避けるため遅延インポート

    d = syntax_dir or _DEFAULT_SYNTAX_DIR
    results = []
    for name in list_syntax_names(syntax_dir):
        filepath = os.path.join(d, f"{name}.json")
        error_info = None
        for filename, code in sources.items():
            try:
                YuiParser(filepath).parse(code)
            except Exception as e:
                error_info = (filename, e)
                break
        if error_info is None:
            results.append((name, True, 'OK'))
        else:
            fname, err = error_info
            results.append((name, False, f"FAIL ({os.path.basename(fname)}: {err})"))
    return results


def generate_bnf(terminals: dict) -> str:
    """
    syntax dict から BNF 風の文法表記を生成する。
    各トークン名には for_example() で得た代表文字列を当てはめる。
    """
    s = YuiSyntax(terminals)
    syntax_name = terminals.get('syntax', '?')

    def ex(name: str) -> str:
        """トークンの代表文字列。未定義なら ''。"""
        if not s.is_defined(name):
            return ''
        val = s.for_example(name)
        # assert-infix が改行の場合は記号に置換
        if val == '\n':
            return '↵'
        return val

    E, N, B = '<expr>', '<name>', '<block>'

    out: list[str] = []

    def section(title: str):
        out.append('')
        out.append(f'{title}:')

    def r(lhs: str, *parts):
        """BNF規則を追加。空文字列の部品は無視する。"""
        tokens = [str(p) for p in parts if p is not None and str(p) != '']
        if tokens:
            out.append(f'  {lhs:<18} ::= {" ".join(tokens)}')

    # ── ヘッダ ────────────────────────────────────────────────
    out.append(f'Grammar for {syntax_name!r}')
    out.append('─' * 40)

    # ── Literals ─────────────────────────────────────────────
    section('Literals')

    r('Number', '0  |  3.14')

    sb, se = ex('string-begin'), ex('string-end')
    ib, ie = ex('string-interpolation-begin'), ex('string-interpolation-end')
    interp = f'  (interp: {ib}{E}{ie})' if ib else ''
    r('String', f'{sb}...{se}{interp}')

    ab, ae = ex('array-begin') or '[', ex('array-end') or ']'
    asep = ex('array-separator') or ','
    r('Array', f'{ab} {E} {{{asep} {E}}} {ae}')

    ob, oe = ex('object-begin') or '{', ex('object-end') or '}'
    kvsep = ex('key-value-separator') or ':'
    osep = ex('object-separator') or ','
    r('Object', f'{ob} "key"{kvsep}{E} {{{osep} "key"{kvsep}{E}}} {oe}')

    nl = ex('null')
    if nl:
        r('Null', nl)

    bt, bf = ex('boolean-true'), ex('boolean-false')
    if bt or bf:
        r('Boolean', bt or '?', '|', bf or '?')

    nb, ne = ex('extra-name-begin'), ex('extra-name-end')
    if nb:
        r('Name', f'{nb}...{ne}  |  letter...')
    else:
        r('Name', 'letter...')

    # ── Expressions ──────────────────────────────────────────
    section('Expressions')

    gb, ge = ex('grouping-begin'), ex('grouping-end')
    if gb:
        r('Grouping', f'{gb} {E} {ge}')

    if ex('length-begin'):
        r('Length', ex('length-begin') + E + ex('length-end'))
    elif ex('unary-length'):
        r('Length', ex('unary-length') + ' ' + E)
    elif ex('property-accessor') and ex('property-length'):
        r('Length', f'{E}{ex("property-accessor")}{ex("property-length")}')

    if ex('unary-minus'):
        r('Minus', ex('unary-minus') + E)

    fa_b, fa_e = ex('funcapp-args-suffix'), ex('funcapp-args-end')
    fa_sep = ex('funcapp-args-separator') or ','
    if fa_b:
        r('FuncApp', f'{E}{fa_b}{E} {{{fa_sep} {E}}}{fa_e}')

    ai_b = ex('array-indexer-suffix') or '['
    ai_e = ex('array-indexer-end') or ']'
    r('Index', f'{E}{ai_b}{E}{ai_e}')

    pa = ex('property-accessor')
    if pa:
        props = [ex(k) for k in ('property-length', 'property-type') if ex(k)]
        if props:
            r('Property', f'{E}{pa}({" | ".join(props)})')

    arith = [(op, ex(f'binary{op}')) for op in ['+', '-', '*', '/', '%'] if ex(f'binary{op}')]
    if arith:
        r('Arithmetic', ' | '.join(f'{E} {t} {E}' for _, t in arith))

    comp_ops = [
        (op, ex(f'binary{op}'))
        for op in ['==', '!=', '<', '<=', '>', '>=', 'in', 'notin']
        if ex(f'binary{op}')
    ]
    if comp_ops:
        r('Comparison', ' | '.join(f'{E} {t} {E}' for _, t in comp_ops))

    # ── Statements ───────────────────────────────────────────
    section('Statements')

    r('Assignment',
      ex('assignment-begin'), E, ex('assignment-infix'), E, ex('assignment-end'))
    r('Increment',
      ex('increment-begin'), E, ex('increment-infix'), ex('increment-end'))
    r('Decrement',
      ex('decrement-begin'), E, ex('decrement-infix'), ex('decrement-end'))
    r('Append',
      ex('append-begin'), E, ex('append-infix'), E, ex('append-suffix'), ex('append-end'))

    if ex('break'):    r('Break',    ex('break'))
    if ex('continue'): r('Continue', ex('continue'))
    if ex('pass'):     r('Pass',     ex('pass'))

    r('Return', ex('return-begin'), E, ex('return-end'))
    if ex('return-none'):
        r('Return (void)', ex('return-none'))

    r('Print', ex('print-begin'), E, ex('print-end'))

    r('Repeat',
      ex('repeat-begin'), E, ex('repeat-times'), ex('repeat-block'), B, ex('repeat-end'))

    # If ─ infix-op形式（pylike/emoji）か suffix形式（yui）かで分岐
    if_b  = ex('if-begin')
    if_cb = ex('if-condition-begin')
    if_ce = ex('if-condition-end')
    if_then = ex('if-then')
    if_else = ex('if-else')
    if_end  = ex('if-end')

    if_infix_ops = [
        (op, ex(f'if-infix{op}'))
        for op in ['==', '!=', '<', '<=', '>', '>=', 'in', 'notin']
        if ex(f'if-infix{op}')
    ]
    if_suffix_ops = [
        ex(f'if-suffix{op}')
        for op in ['!=', '<', '<=', '>', '>=', 'in', 'notin', '==']
        if ex(f'if-suffix{op}')
    ]

    if if_infix_ops:
        op_alts = ' | '.join(f'{E} {t} {E}' for _, t in if_infix_ops)
        r('If', if_b, if_cb, f'({op_alts})', if_ce, if_then, B,
          f'[ {if_else} {B} ]' if if_else else '', if_end)
    else:
        infix_word = ex('if-infix')
        suffix_str = f'[ {" | ".join(if_suffix_ops[:3])}... ]' if if_suffix_ops else ''
        r('If', if_b, if_cb, E, infix_word, E, suffix_str, if_ce, if_then, B,
          f'[ {if_else} {B} ]' if if_else else '', if_end)

    # FuncDef
    fd_b    = ex('funcdef-begin')
    fd_nb   = ex('funcdef-name-begin')
    fd_ne   = ex('funcdef-name-end')
    fd_noarg = ex('funcdef-noarg')
    fd_ab   = ex('funcdef-args-begin')
    fd_ae   = ex('funcdef-args-end')
    fd_asep = ex('funcdef-arg-separator')
    fd_blk  = ex('funcdef-block')
    fd_end  = ex('funcdef-end')

    if fd_noarg:
        args_part = f'( {fd_noarg}  |  {fd_ab} {N} {{{fd_asep} {N}}} {fd_ae} )'
    else:
        args_part = f'{fd_ab} {N} {{{fd_asep} {N}}} {fd_ae}'

    r('FuncDef', fd_b, fd_nb, N, fd_ne, args_part, fd_blk, B, fd_end)

    # Assert
    if s.is_defined('assert-begin'):
        r('Assert', ex('assert-begin'), E, ex('assert-infix'), E, ex('assert-end'))

    if s.is_defined('import-standard'):
        r('Import', ex('import-standard'))

    # ── Blocks ───────────────────────────────────────────────
    section('Blocks')

    blk_b, blk_e = ex('block-begin'), ex('block-end')
    if blk_b and blk_e:
        r('Block', blk_b, '<stmt>...', blk_e)
    elif blk_b:
        r('Block', blk_b, '<stmt>...', '(indent-delimited)')
    else:
        r('Block', '<stmt>...', '(indent-delimited)')
    r('TopLevel', '<stmt>...')

    # ── Comments ─────────────────────────────────────────────
    section('Comments')
    if s.is_defined('line-comment-begin'):
        r('LineComment', ex('line-comment-begin') + ' ...')
    if s.is_defined('comment-begin') and s.is_defined('comment-end'):
        r('BlockComment', ex('comment-begin') + ' ... ' + ex('comment-end'))

    return '\n'.join(out)


class YuiSyntax(object):
    terminals: Dict[str, str]

    def __init__(self, syntax_json: Dict[str, str]):
        super().__init__()
        assert isinstance(syntax_json, dict), "Terminals must be a dictionary"
        self.terminals = syntax_json.copy()

    def is_defined(self, terminal):
        return self.terminals.get(terminal, "") != ""

    def update_syntax(self, **kwargs):
        self.terminals.update(kwargs)

    def get_pattern(self, terminal: str, if_undefined = "") -> re.Pattern:
        pattern = self.terminals.get(terminal, if_undefined)
        if isinstance(pattern, str):
            try:
                pattern = re.compile(pattern)
            except re.error:
                raise ValueError(f"Invalid regex '{terminal}': {pattern}")
            self.terminals[terminal] = pattern
        return pattern

    def for_example(self, terminal: str) -> str:
        if self.is_defined(terminal):
            pattern = self.terminals[terminal]
            if not isinstance(pattern, str):
                pattern = pattern.pattern
            example = get_example_from_pattern(pattern)
            return example
        return ""




# Source
from dataclasses import dataclass
from typing import List, Optional, Dict, Any, Union
import re

from .yuiast import (
    ASTNode,
    ConstNode, NameNode, StringNode, NumberNode, ArrayNode, ObjectNode,
    MinusNode, ArrayLenNode,
    FuncAppNode, GetIndexNode, BinaryNode,
    AssignmentNode, IncrementNode, DecrementNode, AppendNode,
    BlockNode, PrintExpressionNode, PassNode,
    IfNode, BreakNode, RepeatNode, FuncDefNode, ReturnNode,
    AssertNode, ImportNode,
)

from .yuitypes import YuiValue, YuiType, YuiError
from .yuisyntax import YuiSyntax, load_syntax

def is_all_alnum(s: str) -> bool:
    for ch in s:
        if not (('a' <= ch <= 'z') or ('A' <= ch <= 'Z') or ('0' <= ch <= '9') or (ch == '_')):
            return False
    return True

def extract_identifiers(text):
    identifiers = []
    
    # 識別子のパターン: アルファベットまたはアンダースコアで始まり、
    # その後に英数字とアンダースコアが続く
    identifier_pattern = r'[^\s\]\[\(\)"]+'
    
    # 1. 改行と= の間（==は除外）
    pattern1 = rf'\n\s*({identifier_pattern})\s*=(?!=)'
    matches1 = re.findall(pattern1, text)
    identifiers.extend(matches1)
    
    # 2. 関数名のパターン
    pattern2 = rf'({identifier_pattern})\s*[\(]'
    matches2 = re.findall(pattern2, text)
    identifiers.extend(matches2)
    
    def has_unicode(s):
        for ch in s:
            if ord(ch) > 127:
                return True
        return False

    # Unicode文字を含む文字列のみ
    identifiers = list(set(id for id in identifiers if has_unicode(id)))
    # print(f"@Extracted identifiers: {identifiers}")  
    return list(set(identifiers))

@dataclass
class SourceNode(ASTNode):
    """null値（?）を表すノード"""
    def __init__(self):
        super().__init__()

class Source(YuiSyntax):
    """ソースコード"""
    def __init__(self, source: str, filename: str = "main.yui", pos: int = 0, syntax = 'yui'):
        terminals = load_syntax(syntax) if isinstance(syntax, str) else syntax
        super().__init__(terminals)
        self.filename = filename
        self.source = source
        self.pos = pos
        self.length = len(source)
        self.special_names = []
        self.add_special_names(extract_identifiers(source))      
        self.memos = {}
        # self.backtrack = True
    
    def get_memo(self, nonterminal: str, pos: int):
        """Pakrat parsingのメモを取得する"""
        return self.memos.get((nonterminal, pos), None) 
    
    def set_memo(self, nonterminal: str, pos: int, result: Any, new_pos: int):
        """Pakrat parsingのメモを設定する"""
        self.memos[(nonterminal, pos)] = (result, new_pos)

    def has_next(self):
        return self.pos < self.length

    def is_eos(self):
        return self.pos >= self.length

    def consume_string(self, text: str):
        if self.source.startswith(text, self.pos):
            self.pos += len(text)
            return True
        return False
    
    def matched_string(self, terminal: str) -> Optional[str]:
        pattern = self.get_pattern(terminal)
        match_result = pattern.match(self.source, self.pos)
        if match_result:
            return match_result.group(0)
        return ''

    def is_match(self, terminal: str, if_undefined:Union[bool, str]=False, unconsumed=False, lskip_ws=True, lskip_lf=False):
        if isinstance(if_undefined, bool) and not self.is_defined(terminal):
            return if_undefined
        if terminal.startswith('!'): # よくある間違いを検出
            wrong_terminal = f"!{terminal}"
            if self.is_defined(wrong_terminal) and self.is_match(wrong_terminal, unconsumed=True, lskip_ws=lskip_ws):
                expected = self.for_example(terminal)
                matched = self.matched_string(wrong_terminal)
                raise YuiError(("wrong", "token", f"❌`{matched}`", f"✅`{expected}`"), self.p(length=1))
        saved_pos = self.pos    
        if lskip_ws or lskip_lf:
            self.skip_whitespaces_and_comments(include_linefeed=lskip_lf)
        pattern = self.get_pattern(terminal, if_undefined=if_undefined if isinstance(if_undefined, str) else "")
        match_result = pattern.match(self.source, self.pos)
        if match_result:
            match_length = match_result.end() - match_result.start()
            if unconsumed:
                self.pos = saved_pos
                return True
            self.pos += match_length
            return True
        self.pos = saved_pos
        return False
    
    def try_match(self, terminal: str, if_undefined=True, unconsumed=False, 
                  BK=False, lskip_ws=True, lskip_lf=False, opening_pos: int = None):
        if self.is_match(terminal, if_undefined=if_undefined, unconsumed=unconsumed, lskip_ws=lskip_ws, lskip_lf=lskip_lf):
            return
        expected_token = self.for_example(terminal)
        if opening_pos is not None:
            raise YuiError(("expected", "closing", f"✅`{expected_token}`"), self.p(start_pos=opening_pos))
        snippet = self.capture_line()
        raise YuiError(("expected", "token", f"✅`{expected_token}`", f"❌`{snippet}`", f"🔍{terminal}"), self.p(length=1), BK=BK)

    def find_match(self, terminal: str, suffixes: List[str], lskip_lf=False) -> Optional[str]:
        for suffix in suffixes:
            key = f"{terminal}{suffix}"
            if self.is_match(key, if_undefined=False, unconsumed=False, lskip_ws=True, lskip_lf=lskip_lf):
                return suffix
        return None

    def can_backtrack(self, lookahead: List[str]) -> bool:
        if self.is_defined(lookahead):
            pattern = self.get_pattern(lookahead)
            captured = self.capture_line()
            return not pattern.search(captured)
        return True

    def is_eos_or_linefeed(self, lskip_ws=True, unconsumed=False):
        saved_pos = self.pos    
        if lskip_ws:
            self.skip_whitespaces_and_comments(include_linefeed=False)
        if self.is_eos():
            if unconsumed:
                self.pos = saved_pos
            return True
        return self.is_match("linefeed", lskip_ws=False, lskip_lf=False, unconsumed=unconsumed)

    def consume_until(self, terminal:str, until_eof=True, disallow_string:str=None):
        pattern = self.get_pattern(terminal)
        match_result = pattern.search(self.source, self.pos)
        if match_result:
            self.pos = match_result.start()
            if disallow_string:
                if disallow_string in self.source[self.pos:match_result.start()]:
                    return False
            return True
        if until_eof:
            self.pos = self.length
            return True
        return False

    def skip_whitespaces_and_comments(self, include_linefeed=False):
        """空白文字とコメントをスキップする"""
        while self.has_next():
            if self.is_match("whitespaces", lskip_ws=False):
                continue
            if include_linefeed and self.is_match("linefeed", lskip_ws=False):
                continue
            if self.is_match("line-comment-begin", if_undefined=False, lskip_ws=False):
                self.consume_until("linefeed", until_eof=True)
                continue
            if self.is_defined("comment-begin") and self.is_defined("comment-end"):
                opening_pos = self.pos
                if self.is_match("comment-begin", lskip_ws=False):
                    self.consume_until("comment-end", until_eof=True)
                    self.try_match("comment-end", lskip_ws=False, opening_pos=opening_pos)
                    continue
            break

    def add_special_names(self, names: List[str]):
        special_names = set(self.special_names + names)
        self.special_names = sorted(special_names, key=len, reverse=True)

    def match_special_name(self, unconsumed=False) -> Optional[str]:
        for name in self.special_names:
            if self.source.startswith(name, self.pos):
                if not unconsumed:
                    self.pos += len(name)
                return name
        return None

    def capture_indent(self, indent_chars: str = " \t　"):
        start_pos = self.pos - 1
        # 行の先頭位置を探す
        while 0 <= start_pos:
            char = self.source[start_pos]
            if char == '\n':
                start_pos += 1
                break
            start_pos -= 1
        end_pos = start_pos
        while end_pos < self.length:
            char = self.source[end_pos]
            if char in indent_chars:
                end_pos += 1
            else:
                break
        return self.source[start_pos:end_pos]

    def capture_line(self):
        start_pos = self.pos
        while self.pos < self.length:
            if self.is_match("linefeed", lskip_ws=False, unconsumed=True) or \
                 self.is_match("line-comment-begin", lskip_ws=False, unconsumed=True) or \
                self.is_match("comment-begin", lskip_ws=False, unconsumed=True) or \
                self.is_match("statement-separator", lskip_ws=False, unconsumed=True):
                captured = self.source[start_pos:self.pos]
                self.pos = start_pos
                return captured
            self.pos += 1
        self.pos = start_pos
        return self.source[start_pos:].split('\n')[0]

    def capture_comment(self):
        """コメントを位置を変えずに習得する。行コメントとブロックコメントの両方に対応。"""
        save_pos = self.pos
        self.is_match("whitespaces", if_undefined=r"[ \t\r]+")
        comment = None
        if self.is_match("line-comment-begin", if_undefined=False):
            start_pos = self.pos
            self.consume_until("linefeed", until_eof=True)
            comment =self.source[start_pos:self.pos]
        if self.is_match("comment-begin", if_undefined=False):
            start_pos = self.pos
            self.consume_until("comment-end", until_eof=True)
            comment = self.source[start_pos:self.pos]
        self.pos = save_pos # コメントは位置を変えずに取得
        return comment

    def p(self, node: ASTNode = None, start_pos: int = None, end_pos: int = None, length: int = 0) -> ASTNode:
        node = node or SourceNode()
        node.filename = self.filename
        node.source = self.source
        ### FIXME: node.comment = self.capture_comment()

        save_pos = self.pos
        if start_pos is not None:
            node.pos = start_pos
            if end_pos is not None:
                node.end_pos = end_pos
            elif length != 0:
                node.end_pos = min(start_pos + length, self.length)
            else:
                node.end_pos = save_pos           
        elif length != 0:
            node.pos = self.pos
            node.end_pos = min(self.pos + length, self.length)
        else:
            node.pos = max(self.pos-1, 0)
            node.end_pos = self.pos
        return node

NONTERMINALS = {}

class ParserCombinator(object):

    def quick_check(self, source: Source) -> bool:
        return True
    
    def match(self, source: Source, pc: dict):
        return True

def parse(nonterminal: str, source: Source, pc: dict, lskip_ws=True, lskip_lf=False, BK=False) -> Any:
    global NONTERMINALS
    patterns = NONTERMINALS[nonterminal]
    saved_pos = source.pos
    if lskip_ws or lskip_lf:
        source.skip_whitespaces_and_comments(include_linefeed=lskip_lf)
    
    memo = source.get_memo(nonterminal, source.pos)
    if memo is not None:
        source.pos = memo[1]
        result = memo[0]
    else:
        saved_pos = source.pos
        try:
            result = patterns.match(source, pc)
            source.set_memo(nonterminal, saved_pos, result, source.pos)
        except YuiError as e:
            #print(f"@fail {nonterminal} BK={e.BK} {e}")
            if e.BK == True and BK == False:
                source.pos = saved_pos
                snippet = source.capture_line()
                raise YuiError(("expected", nonterminal[1:].lower(), f"❌{snippet}", f"⚠️{e}"), source.p(length=1))
            raise e
    return result

def is_parsable(nonterminal: str, source: Source, pc: dict, lskip_ws=True, lskip_lf=False) -> bool:
    try:
        parse(nonterminal, source, pc, lskip_ws=lskip_ws, lskip_lf=lskip_lf)
        return True
    except YuiError:
        return False

class ConstParser(ParserCombinator):

    def quick_check(self, source: Source) -> bool:
        return source.is_match("null", if_undefined=False) or \
            source.is_match("boolean-true", if_undefined=False) or \
            source.is_match("boolean-false", if_undefined=False)

    def match(self, source: Source, pc: dict):
        saved_pos = source.pos
        if source.is_match("null", if_undefined=False):
            return source.p(ConstNode(None), start_pos=saved_pos)
        if source.is_match("boolean-true", if_undefined=False):
            return source.p(ConstNode(True), start_pos=saved_pos)
        if source.is_match("boolean-false", if_undefined=False):
            return source.p(ConstNode(False), start_pos=saved_pos)
        raise YuiError(("expected", "null or boolean"), source.p(length=1), BK=True)

NONTERMINALS["@Boolean"] = ConstParser()

class NumberParser(ParserCombinator):

    def quick_check(self, source: Source) -> bool:
        return source.is_match("number-first-char", unconsumed=True)
    
    def match(self, source: Source, pc: dict):
        saved_pos = source.pos
        if source.is_match("number-first-char"):
            source.try_match("number-chars", lskip_ws=False)
            if source.is_match("number-dot-char", lskip_ws=False):
                source.try_match("number-first-char", lskip_ws=False)
                source.try_match("number-chars", lskip_ws=False)
                number = source.source[saved_pos:source.pos]
                return source.p(NumberNode(float(number)), start_pos=saved_pos)
            else:
                number = source.source[saved_pos:source.pos]
            return source.p(NumberNode(int(number)), start_pos=saved_pos)
        raise YuiError(("expected", "number"), source.p(length=1), BK=True)

NONTERMINALS["@Number"] = NumberParser()

class StringParser(ParserCombinator):

    def quick_check(self, source: Source) -> bool:
        return source.is_match("string-begin", unconsumed=True)
    
    def match(self, source: Source, pc: dict):
        opening_quote_pos = source.pos
        if source.is_match("string-begin"):
            opening_pos = source.pos
            string_content = []
            expression_count = 0
            while source.pos < source.length:
                source.consume_until("string-content-end", until_eof=True)
                string_content.append(source.source[opening_pos:source.pos])
                if source.is_match("string-end", unconsumed=True):
                    break
                if source.is_match("string-escape"):
                    if source.is_eos():
                        raise YuiError(("bad", "escape sequence"), source.p(length=1))
                    next_char = source.source[source.pos]
                    source.pos += 1
                    if next_char == 'n':
                        string_content.append('\n')
                    elif next_char == 't':
                        string_content.append('\t')
                    else:
                        string_content.append(next_char)
                    continue
                start_inter_pos = source.pos
                if source.is_match("string-interpolation-begin", lskip_ws=False):
                    expression = parse("@Expression", source, pc)
                    source.try_match("string-interpolation-end", opening_pos=start_inter_pos)
                    string_content.append(expression)
                    expression_count += 1
                    continue
            source.try_match("string-end", lskip_ws=False, opening_pos=opening_quote_pos)
            if expression_count == 0:
                string_content = ''.join(string_content)
            return source.p(StringNode(string_content), start_pos=opening_quote_pos)
        raise YuiError(("expected", "string"), source.p(length=1), BK=True)

NONTERMINALS["@String"] = StringParser()

class ArrayParser(ParserCombinator):

    def quick_check(self, source: Source) -> bool:
        return source.is_match("array-begin", unconsumed=True)
    
    def match(self, source: Source, pc: dict):
        opening_pos = source.pos
        if source.is_match("array-begin"):
            arguments = []
            while not source.is_match("array-end", lskip_lf=True, unconsumed=True):
                arguments.append(parse("@Expression", source, pc, lskip_lf=True))
                if source.is_match("array-separator", lskip_lf=True):
                    continue
            source.try_match("array-end", lskip_lf=True, opening_pos=opening_pos)
            return source.p(ArrayNode(arguments), start_pos=opening_pos)
        raise YuiError(("expected", "array"), source.p(length=1), BK=True)

NONTERMINALS["@Array"] = ArrayParser()

class ObjectParser(ParserCombinator):
    def quick_check(self, source: Source) -> bool:
        return source.is_match("object-begin", unconsumed=True)
    
    def match(self, source: Source, pc: dict):
        opening_pos = source.pos
        if source.is_match("object-begin", lskip_lf=True):
            arguments = []
            while not source.is_match("object-end", lskip_lf=True, unconsumed=True):
                arguments.append(parse("@String", source, pc, lskip_lf=True))
                source.try_match("key-value-separator", lskip_lf=True)
                arguments.append(parse("@Expression", source, pc, lskip_lf=True))
                if source.is_match("object-separator", lskip_lf=True):
                    continue
            source.try_match("object-end", lskip_lf=True, opening_pos=opening_pos)
            return source.p(ObjectNode(arguments), start_pos=opening_pos)
        raise YuiError(("expected", "object"), source.p(length=1), BK=True)

NONTERMINALS["@Object"] = ObjectParser()

class NameParser(ParserCombinator):

    def match(self, source: Source, pc: dict):
        special_name = source.match_special_name()
        if special_name is not None:
            return source.p(NameNode(special_name), start_pos=source.pos-len(special_name))
        if source.is_match("extra-name-begin", if_undefined=False):
            start_pos = source.pos
            source.consume_until("extra-name-end", disallow_string="\n")
            name = source.source[start_pos:source.pos]
            node = source.p(NameNode(name), start_pos=start_pos)
            source.try_match("extra-name-end", opening_pos=start_pos-1)
            return node
        start_pos = source.pos
        if source.is_match("name-first-char"):
            source.try_match("name-chars", lskip_ws=False)
            source.try_match("name-end", lskip_ws=False)
            name = source.source[start_pos:source.pos]
            return source.p(NameNode(name), start_pos=start_pos)
        snippet = source.capture_line().strip()
        raise YuiError(("wrong", "name", f"❌{snippet}"), source.p(length=1), BK=True)

NONTERMINALS["@Name"] = NameParser()

LITERALS = ["@Number","@String","@Array","@Object","@Boolean"]

class TermParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        opening_pos = source.pos
        if source.is_match("grouping-begin"):
            expression_node = parse("@Expression", source, pc)
            source.try_match("grouping-end", opening_pos=opening_pos)
            return expression_node
        if source.is_match("length-begin"):
            expression_node = parse("@Expression", source, pc)
            source.try_match("length-end", opening_pos=opening_pos)
            return source.p(ArrayLenNode(expression_node), start_pos=opening_pos)
        for literal in LITERALS:
            if NONTERMINALS[literal].quick_check(source):
                source.pos = opening_pos
                return parse(literal, source, pc, BK=True)
        return parse("@Name", source, pc, BK=True)

NONTERMINALS["@Term"] = TermParser()

class PrimaryParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        start_pos = source.pos
        if source.is_match("unary-minus", if_undefined=False):
            node = parse('@Primary', source, pc)
            return source.p(MinusNode(node), start_pos=start_pos)
        if source.is_match('unary-inspection', if_undefined=False):
            node = parse('@Primary', source, pc)
            return source.p(PrintExpressionNode(node, inspection=True), start_pos=start_pos)
        if source.is_match('unary-length', if_undefined=False):
            node = parse('@Primary', source, pc)
            return source.p(ArrayLenNode(node), start_pos=start_pos)
        node = parse("@Term", source, pc, BK=True)
        while source.has_next():
            opening_pos = source.pos
            if source.is_match("funcapp-args-suffix"):
                arguments = []
                while not source.is_match("funcapp-args-end", unconsumed=True):
                    arguments.append(parse("@Expression", source, pc, lskip_lf=True))
                    if source.is_match("funcapp-args-separator"):
                        continue
                    #break
                source.try_match("funcapp-args-end", opening_pos=opening_pos)
                node = source.p(FuncAppNode(node, arguments), start_pos=start_pos)
                continue
            if source.is_match("array-indexer-suffix", if_undefined=r"\["):
                index_node = parse("@Expression", source, pc)
                source.try_match("array-indexer-end", opening_pos=opening_pos)
                node = source.p(GetIndexNode(node, index_node), start_pos=start_pos)
                continue
            save_pos = source.pos
            if source.is_match("property-accessor", if_undefined=False):
                if source.is_match("property-length", if_undefined=False):
                    node = source.p(ArrayLenNode(node), start_pos=start_pos)
                    continue
                if source.is_match("property-type", if_undefined=False):
                    ...
                source.pos = save_pos
            break
        return node
    
NONTERMINALS["@Primary"] = PrimaryParser()

class MultiplicativeParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        start_pos = source.pos
        left_node = parse("@Primary", source, pc, BK=True)
        try:
            if source.is_match("binary*", if_undefined=False):
                right_node = parse("@Multiplicative", source, pc)
                return source.p(BinaryNode(left_node, "*", right_node), start_pos=start_pos)
            if source.is_match("binary/", if_undefined=False):
                right_node = parse("@Multiplicative", source, pc)
                return source.p(BinaryNode(left_node, "/", right_node), start_pos=start_pos)
            if source.is_match("binary%", if_undefined=False):
                right_node = parse("@Multiplicative", source, pc)
                return source.p(BinaryNode(left_node, "%", right_node), start_pos=start_pos)
        except YuiError:
            pass
        source.pos = left_node.end_pos
        return left_node

NONTERMINALS["@Multiplicative"] = MultiplicativeParser()

class AdditiveParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        start_pos = source.pos
        left_node = parse("@Multiplicative", source, pc, BK=True)
        try:
            if source.is_match("binary+", if_undefined=False):
                right_node = parse("@Additive", source, pc)
                return source.p(BinaryNode(left_node, "+", right_node), start_pos=start_pos)
            if source.is_match("binary-", if_undefined=False):
                right_node = parse("@Additive", source, pc)
                return source.p(BinaryNode(left_node, "-", right_node), start_pos=start_pos)
        except YuiError:
            pass
        source.pos = left_node.end_pos
        return left_node

NONTERMINALS["@Additive"] = AdditiveParser()

class ComparativeParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        start_pos = source.pos
        left_node = parse("@Additive", source, pc, BK=True)
        try:
            if source.is_match("binary==", if_undefined=False):
                right_node = parse("@Additive", source, pc)
                return source.p(BinaryNode(left_node, "==", right_node, comparative=True), start_pos=start_pos)
            if source.is_match("binary!=", if_undefined=False):
                right_node = parse("@Additive", source, pc)
                return source.p(BinaryNode(left_node, "!=", right_node, comparative=True), start_pos=start_pos)
            if source.is_match("binary<=", if_undefined=False):
                right_node = parse("@Additive", source, pc)
                return source.p(BinaryNode(left_node, "<=", right_node, comparative=True), start_pos=start_pos)
            if source.is_match("binary>=", if_undefined=False):
                right_node = parse("@Additive", source, pc)
                return source.p(BinaryNode(left_node, ">=", right_node, comparative=True), start_pos=start_pos)
            if source.is_match("binary<", if_undefined=False):
                right_node = parse("@Additive", source, pc)
                return source.p(BinaryNode(left_node, "<", right_node, comparative=True), start_pos=start_pos)
            if source.is_match("binary>", if_undefined=False):
                right_node = parse("@Additive", source, pc)
                return source.p(BinaryNode(left_node, ">", right_node, comparative=True), start_pos=start_pos)
            if source.is_match("binaryin", if_undefined=False):
                right_node = parse("@Additive", source, pc)
                return source.p(BinaryNode(left_node, "in", right_node, comparative=True), start_pos=start_pos)
            if source.is_match("binarynotin", if_undefined=False):
                right_node = parse("@Additive", source, pc)
                return source.p(BinaryNode(left_node, "notin", right_node, comparative=True), start_pos=start_pos)
        except YuiError:
            pass
        source.pos = left_node.end_pos
        return left_node

NONTERMINALS["@Comparative"] = ComparativeParser()

class ExpressionParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        return parse("@Comparative", source, pc, BK=True)

NONTERMINALS["@Expression"] = ExpressionParser()

## Statement

class AssignmentParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        BK = source.can_backtrack('assignment-lookahead')
        start_pos = source.pos
        source.try_match('assignment-begin', BK=BK)
        if BK: BK = source.pos == start_pos
        left_node = parse("@Expression", source, pc, BK=BK)
        source.try_match('assignment-infix', BK=BK)
        right_node = parse("@Expression", source, pc, BK=BK)
        source.try_match('assignment-end', BK=BK) # E E 代入もありえる
        return source.p(AssignmentNode(left_node, right_node), start_pos=start_pos)
    
NONTERMINALS["@Assignment"] = AssignmentParser()

class IncrementParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        BK = source.can_backtrack('increment-lookahead')
        start_pos = source.pos
        source.try_match('increment-begin', BK=BK)
        if BK: BK = source.pos == start_pos

        lvalue_node = parse("@Expression", source, pc, BK=BK)
        source.try_match('increment-infix', BK=BK)
        source.try_match('increment-end', BK=BK)
        return source.p(IncrementNode(lvalue_node), start_pos=start_pos)

NONTERMINALS["@Increment"] = IncrementParser()

class DecrementParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        BK = source.can_backtrack('decrement-lookahead')
        start_pos = source.pos
        source.try_match('decrement-begin', BK=BK)
        if BK: BK = source.pos == start_pos
        lvalue_node = parse("@Expression", source, pc, BK=BK)
        source.try_match('decrement-infix', BK=BK)
        source.try_match('decrement-end', BK=BK)
        return source.p(DecrementNode(lvalue_node), start_pos=start_pos)

NONTERMINALS["@Decrement"] = DecrementParser()

class AppendParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        BK = source.can_backtrack('append-lookahead')
        start_pos = source.pos
        source.try_match('append-begin', BK=BK)
        if BK: BK = source.pos == start_pos
        lvalue_node = parse("@Expression", source, pc, BK=BK)
        source.try_match('append-infix', BK=BK)
        value = parse("@Expression", source, pc, BK=BK)
        source.try_match('append-suffix', BK=BK)
        source.try_match('append-end', BK=BK)
        return source.p(AppendNode(lvalue_node, value), start_pos=start_pos)

NONTERMINALS["@Append"] = AppendParser()

class BreakParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        start_pos = source.pos
        source.try_match('break', BK=True)
        return source.p(BreakNode(), start_pos=start_pos)

NONTERMINALS["@Break"] = BreakParser()

class ImportParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        start_pos = source.pos
        source.try_match('import-standard', BK=True)
        return source.p(ImportNode(), start_pos=start_pos)

NONTERMINALS["@Import"] = ImportParser()


class PassParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        start_pos = source.pos
        source.try_match('pass', BK=True)
        return source.p(PassNode(), start_pos=start_pos)
    
NONTERMINALS["@Pass"] = PassParser()

class ReturnParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        BK = source.can_backtrack('return-lookahead')
        start_pos = source.pos
        source.try_match('return-begin', BK=BK)
        if BK: BK = source.pos == start_pos
        expr_node = parse("@Expression", source, pc, BK=BK)
        source.try_match('return-end', BK=BK)
        return source.p(ReturnNode(expr_node), start_pos=start_pos)
    
NONTERMINALS["@Return"] = ReturnParser()

class PrintExpressionParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        BK = source.can_backtrack('print-lookahead')
        start_pos = source.pos
        source.try_match('print-begin', BK=BK)
        if BK: BK = source.pos == start_pos
        expr_node = parse("@Expression", source, pc, BK=BK)
        source.try_match('print-end', BK=BK)
        return source.p(PrintExpressionNode(expr_node), start_pos=start_pos)

NONTERMINALS["@PrintExpression"] = PrintExpressionParser()

class RepeatParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        BK = source.can_backtrack('repeat-lookahead')
        start_pos = source.pos
        source.try_match('repeat-begin', BK=BK)
        if BK: BK = source.pos == start_pos
        times_node = parse("@Expression", source, pc, BK=BK)
        source.try_match('repeat-times', BK=BK)
        source.try_match('repeat-block', BK=BK)
        pc = pc.copy()
        pc['indent'] = source.capture_indent()
        block_node = parse("@Block", source, pc)
        source.try_match('repeat-end', BK=False)
        return source.p(RepeatNode(times_node, block_node), start_pos=start_pos)

NONTERMINALS["@Repeat"] = RepeatParser()

class IfParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        start_pos = source.pos
        BK = source.can_backtrack('if-lookahead')
        source.try_match('if-begin', BK=BK) # もし
        source.try_match('if-condition-begin', BK=BK)
        if BK: BK = source.pos == start_pos
        left_node = parse("@Expression", source, pc, BK=BK)
        if isinstance(left_node, BinaryNode) and left_node.comparative:
            operator = left_node.operator
            right_node = left_node.right
            left_node = left_node.left
        else:
            operator = source.find_match('if-infix', ['==', '!=', '<=', '<', '>=', '>', 'notin', 'in'])
            if not operator:
                source.try_match('if-infix', BK=BK) # が
            right_node = parse("@Expression", source, pc, BK=BK)
            
            if not operator:
                operator = source.find_match('if-suffix', ['!=', '<=', '<', '>=', '>', 'notin', 'in', '=='])
                source.skip_whitespaces_and_comments()
                if operator is None:
                    operator = "=="

        source.try_match('if-condition-end', BK=BK) # )

        pc = pc.copy()
        pc['indent'] = source.capture_indent()
        source.try_match('if-then', BK=False) # ならば                

        then_node = parse("@Block", source, pc, BK=False)

        save_pos = source.pos
        source.skip_whitespaces_and_comments(include_linefeed=True)
        if source.is_match('if-else'):
            else_node = parse("@Block", source, pc, BK=False)
        else:
            source.pos = save_pos
            else_node = None
        source.try_match('if-end', BK=False)
        return source.p(IfNode(left_node, operator, right_node, then_node, else_node), start_pos=start_pos)

NONTERMINALS["@If"] = IfParser()

class FuncDefParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        BK = source.can_backtrack('funcdef-lookahead')
        start_pos = source.pos
        source.try_match('funcdef-begin', BK=BK) # もし
        source.try_match('funcdef-name-begin', BK=BK) #
        if BK: BK = source.pos == start_pos
        name_node = parse("@Name", source, pc, BK=BK)
        source.try_match('funcdef-name-end', BK=BK) # =

        arguments = []
        if not source.is_match('funcdef-noarg'): # 引数なし
            source.try_match('funcdef-args-begin', BK=BK) # 入力
            while not source.is_match('funcdef-args-end', unconsumed=True):
                arg_node = parse("@Name", source, pc)
                arguments.append(arg_node)
                if source.is_match('funcdef-arg-separator'):
                    continue
                break
            source.try_match('funcdef-args-end', BK=False)

        pc = pc.copy()
        pc['indent'] = source.capture_indent()
        source.try_match('funcdef-block', BK=BK) # に対し
        body_node = parse("@Block", source, pc, BK=False)
        source.try_match('funcdef-end', BK=False)
        return source.p(FuncDefNode(name_node, arguments, body_node), start_pos=start_pos)

NONTERMINALS["@FuncDef"] = FuncDefParser()

class AssertParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        start_pos = source.pos
        BK = source.can_backtrack('assert-lookahead')
        source.try_match('assert-begin', BK=BK)
        if BK: BK = source.pos == start_pos
        test_node = parse("@Expression", source, pc, BK=BK)
        source.try_match('assert-infix', BK=BK)
        reference_node = parse("@Expression", source, pc, BK=BK)
        source.try_match('assert-end', BK=BK)
        return source.p(AssertNode(test_node, reference_node), start_pos=start_pos)

NONTERMINALS["@Assert"] = AssertParser()

class BlockParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        saved_pos = source.pos
        source.try_match("block-begin")
        if source.is_match('block-end'): # 単一ブロックを認めるか { statement }
            return source.p(BlockNode([]), start_pos=saved_pos)
        
        statements = parse("@Statement[]", source, pc)
        if source.is_match('block-end'):
            return source.p(BlockNode(statements), start_pos=saved_pos)

        end_level_indent = pc.get('indent', '')
        #print(f"@Block in indent='{len(end_level_indent)}'", source.pos, source.capture_line())
        while source.has_next():
            source.is_eos_or_linefeed()
            linestart_pos = source.pos
            if source.consume_string(end_level_indent): 
                if source.is_match('whitespace', lskip_ws=False): # deeper end_level_indent
                    if source.is_match("block-end", unconsumed=True):
                        raise YuiError(("wrong", "indent", f"✅`{end_level_indent}`"), source.p(start_pos=linestart_pos, length=len(end_level_indent)))
                    #print('>>>', source.pos, source.capture_line())
                    statements.extend(parse("@Statement[]", source, pc))
                    #print('<<<', source.pos, source.capture_line())
                    continue
            if source.is_match("linefeed", unconsumed=True):
                continue # ただの空行はブロックの終わりにしない              
            break
        source.try_match("block-end", opening_pos=saved_pos)
        return source.p(BlockNode(statements), start_pos=saved_pos)

NONTERMINALS["@Block"] = BlockParser()

STATEMENTS = [
    "@FuncDef",
    "@Assignment",
    "@Assert",
    "@If",
    "@Repeat",
    "@Break",
    "@Increment",
    "@Decrement",
    "@Append",
    "@Return",
    "@Import",
    "@Pass",
    "@PrintExpression",
]

class StatementParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        saved_pos = source.pos
        #print("@Statement", source.pos, source.capture_line())
        for parser_name in STATEMENTS:
            source.pos = saved_pos # for backtrack
            try:
                statement = parse(parser_name, source, pc, BK=True)
                return statement
            except YuiError as e:
                if e.BK: continue
                raise e
        source.pos = saved_pos
        line = source.capture_line()
        raise YuiError(("wrong", "statement", f"❌{line}"), source.p(length=1))

NONTERMINALS["@Statement"] = StatementParser()

class StatementsParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        if source.is_eos_or_linefeed(lskip_ws=True, unconsumed=True):
            statements = []
        else:
            statements = [parse("@Statement", source, pc)]
        while source.is_match('statement-separator', if_undefined=False):
            statements.append(parse("@Statement", source, pc))
        return statements

NONTERMINALS["@Statement[]"] = StatementsParser()

class TopLevelParser(ParserCombinator):
    def match(self, source: Source, pc: dict):
        source.skip_whitespaces_and_comments(include_linefeed=True)
        saved_pos = source.pos
        statements = []
        while source.has_next():
            cur_pos = source.pos
            statements.extend(parse("@Statement[]", source, pc))
            if cur_pos == source.pos:                
                break
            source.skip_whitespaces_and_comments(include_linefeed=True)
        return source.p(BlockNode(statements, top_level=True), start_pos=saved_pos)

NONTERMINALS["@TopLevel"] = TopLevelParser()

class YuiParser:
    def __init__(self, syntax: Union[str, dict]):
        self.syntax = syntax
        if isinstance(syntax, dict):
            self.terminals = syntax.copy()
        else:
            self.terminals = load_syntax(syntax)
        self.pc = {}
    
    def parse(self, source_code: str) -> ASTNode:
        source = Source(source_code, syntax=self.terminals)
        return parse("@TopLevel", source, {})

